import pytest
import yaml

### Fixtures ###
@pytest.fixture(name="get_default_graph_fixture", scope="session")
def get_default_graph_fixture(client):
    graphfile = client.get_default_graph(channelId=client.channel, filepath="default.yaml")
    with open(graphfile, 'r') as yamlfile:
        graph = yaml.safe_load(yamlfile)
    return graph

@pytest.fixture(name="upload_graph_fixture", scope="session")
def upload_graph_fixture(client, get_default_graph_fixture):
    graph = get_default_graph_fixture
    graphId = client.upload_graph(name="anatools-graph", graph=graph, channelId=client.channel, workspaceId=client.workspace)
    return graphId


### Tests ###
@pytest.mark.smoke
def test_get_graphs(client):
    graphs = client.get_graphs(workspaceId=client.workspace)
    assert isinstance(graphs, list)

@pytest.mark.smoke
def test_get_default_graph(get_default_graph_fixture):
    graph = get_default_graph_fixture
    assert isinstance(graph, dict)
    assert 'nodes' in graph.keys()

@pytest.mark.smoke
def test_upload_graph(upload_graph_fixture):
    graphId = upload_graph_fixture
    assert isinstance(graphId, str)

def test_edit_graph(client, upload_graph_fixture):
    graphId = upload_graph_fixture
    status = client.edit_graph(graphId=graphId, name="anatools-graph-edit", description="anatools test graph edit")
    assert status == True

@pytest.mark.dependency(depends=["test_edit_graph", "test_download_graph"])
def test_delete_graph(client, upload_graph_fixture):
    graphId = upload_graph_fixture
    status = client.delete_graph(graphId=graphId, workspaceId=client.workspace)
    assert status == True

def test_download_graph(client, upload_graph_fixture):
    graphId = upload_graph_fixture
    graph = client.download_graph(graphId=graphId, workspaceId=client.workspace)
    assert isinstance(graph, dict)
    assert 'nodes' in graph.keys()

# def test_set_default_graph(client, create_channel_fixture, upload_graph_fixture):
#     channelId = create_channel_fixture
#     graphId = upload_graph_fixture
#     status = client.set_default_graph(channelId=channelId, graphId=graphId)
#     assert status == True


import os
import pytest

### Fixtures ###
@pytest.fixture(scope="session")
def create_volume_fixture(client):
    volumeId = client.create_volume(name=f"anatools-volume-{client.timestamp}", organizationId=client.organization)
    client.edit_workspace(workspaceId=client.workspace, volumeIds=[volumeId])
    return volumeId

@pytest.fixture(scope="session")
def upload_volume_data_fixture(client, create_volume_fixture):
    volumeId = create_volume_fixture
    file = "test.txt"
    with open(file, "w") as f: f.write("test")
    client.upload_volume_data(volumeId=volumeId, files=[file])
    return file


### Tests ###
def test_get_volumes(client):
    volumes = client.get_volumes()
    assert isinstance(volumes, list)

def test_create_volume(client, create_volume_fixture):
    volumes = client.get_volumes(workspaceId=client.workspace)
    assert isinstance(volumes, list)
    assert len(volumes) == 1
    assert volumes[0]['volumeId'] == create_volume_fixture

@pytest.mark.dependency(depends=["test_delete_volume_data", "test_delete_inpaint"])
def test_delete_volume(client, create_volume_fixture):
    volumeId = create_volume_fixture
    status = client.delete_volume(volumeId=volumeId)
    assert status == True

def test_edit_volume(client, create_volume_fixture):
    volumeId = create_volume_fixture
    status = client.edit_volume(volumeId=volumeId, name=f"anatools-volume-edited-{client.timestamp}")
    assert status == True

def test_upload_volume_data(client, create_volume_fixture, upload_volume_data_fixture):
    volumeId = create_volume_fixture
    file = upload_volume_data_fixture
    volumedata = client.get_volume_data(volumeId=volumeId, files=[file])
    assert isinstance(volumedata, list)
    assert len(volumedata) == 1
    assert volumedata[0]['filename'] == file

def test_download_volume_data(client, create_volume_fixture, upload_volume_data_fixture):
    volumeId = create_volume_fixture
    file = upload_volume_data_fixture
    filepath = client.download_volume_data(volumeId=volumeId, files=[file])
    assert os.path.exists(filepath)  

@pytest.mark.dependency(depends=["test_upload_volume_data","test_download_volume_data"])
def test_delete_volume_data(client, create_volume_fixture, upload_volume_data_fixture):
    volumeId = create_volume_fixture
    file = upload_volume_data_fixture
    status = client.delete_volume_data(volumeId=volumeId, files=[file])
    assert status == True

def test_mount_volumes(client, create_volume_fixture):
    volumeId = create_volume_fixture
    credentials = client.mount_volumes(volumes=[volumeId])
    assert isinstance(credentials, dict)



import os
import json
from PIL import Image, ImageDraw

exts = Image.registered_extensions()
supported_extensions = {ex for ex, f in exts.items() if f in Image.OPEN}

def convert_sagemaker_od(datadir, outdir, mapping = None):
    """ Generate annotations for AWS Sagemaker. Annotation jpegs will be placed in <datadir>/<outputdir>.
    Parameters
    ----------
    datadir : str
        Location of Rendered.ai dataset output.
    outputdir : str
        Name of directory where the results should be written.
    mapfile: str
        The map file used for annotations (YAML only).
    
    Returns
    -------
    """

    annsdir = os.path.join(datadir, "annotations")
    metadir = os.path.join(datadir, "metadata")

    # Get the image shape
    try:
        sample_image_filename = [image for image in os.listdir(datadir + '/images') if os.path.splitext(image)[1].lower() in supported_extensions]
        sample_image = Image.open(datadir + '/images/' + sample_image_filename[0])
        imgshape = [sample_image.size[0], sample_image.size[1], len(sample_image.getbands())]
    except: raise Exception('Could not find a supported image in the dataset images directory.')

    sodcats = list()
    for annsfilename in os.listdir(annsdir):
        with open(os.path.join(annsdir, annsfilename), 'r') as af:
            anns = json.load(af)
        with open(os.path.join(metadir, annsfilename.replace('ana', 'metadata')), 'r') as mf:
            metadata = json.load(mf)

        soddata = dict()
        soddata['file'] = anns['filename']
        soddata['image_size'] = [{'width':imgshape[0],'height':imgshape[1],'depth':imgshape[2]}]
        soddata['annotations'] = list()
        for obj in metadata['objects']:
            if mapping is None:
                for ann in anns['annotations']:
                    if ann['id'] == obj['id']:
                        cat_name = obj['type']  
                        if cat_name not in sodcats: sodcats.append(cat_name)     
                        soddata['annotations'].append({
                            'class_id': sodcats.index(cat_name),
                            'left':     ann['bbox'][0],
                            'top':      ann['bbox'][1],
                            'width':    ann['bbox'][2],
                            'height':   ann['bbox'][3]
                        })
                        break

            else:
                for prop in mapping['properties']:
                    if eval(prop):
                        for ann in anns['annotations']:
                            if ann['id'] == obj['id']:
                                objann = ann
                                break
                        else:  # All the objects from the scene are recorded in metadata; only those in the image are annotated
                            continue

                        rai_cat_id = mapping['properties'][prop]
                        cat = mapping['classes'][rai_cat_id]
                        cat_name = cat[-1]
                        if cat_name not in sodcats: sodcats.append(cat_name)              
                        soddata['annotations'].append({
                            'class_id': sodcats.index(cat_name),
                            'left':     objann['bbox'][0],
                            'top':      objann['bbox'][1],
                            'width':    objann['bbox'][2],
                            'height':   objann['bbox'][3]
                        })
                        break

        soddata['categories'] = list()
        for cId, cName in enumerate(sodcats):
            soddata['categories'].append({'class_id':cId, 'name':cName})
    
        outfile = os.path.join(outdir, '{}.json'.format(anns['filename'].split('.')[0]))
        with open(outfile, 'w') as f:
            json.dump(soddata, f)


def convert_sagemaker_ss(datadir, outdir, mapping=None):
    """ Generate masks for AWS Sagemaker Semantic Segmentation. Mask pngs will be placed in <datadir>/<outputdir>.
    Parameters
    ----------
    datadir : str
        Location of Rendered.ai dataset output.
    outputdir : str
        Name of directory where the results should be written.
    mapfile: str
        The map file used for annotations (YAML only).
    
    Returns
    -------
    """

    annsdir = os.path.join(datadir, "annotations")
    metadir = os.path.join(datadir, "metadata")

    # Get the image shape
    try:
        sample_image_filename = [image for image in os.listdir(datadir + '/images') if os.path.splitext(image)[1].lower() in supported_extensions]
        sample_image = Image.open(datadir + '/images/' + sample_image_filename[0])
        imgshape = [sample_image.size[0], sample_image.size[1], len(sample_image.getbands())]
    except: raise Exception('Could not find a supported image in the dataset images directory.')

    sodcats = ['background']
    for annsfilename in os.listdir(annsdir):
        with open(os.path.join(annsdir, annsfilename), 'r') as af:
            anns = json.load(af)
        with open(os.path.join(metadir, annsfilename.replace('ana', 'metadata')), 'r') as mf:
            metadata = json.load(mf)
        maskimg = Image.new("L", (imgshape[0], imgshape[1]))
        draw = ImageDraw.Draw(maskimg)

        for obj in metadata['objects']:
            if mapping is None:
                for ann in anns['annotations']:
                    if ann['id'] == obj['id']:
                        cat_name = obj['type']  
                        if cat_name not in sodcats: sodcats.append(cat_name)    
                        rai_cat_id = sodcats.index(cat_name)
                        draw.polygon(ann['segmentation'][0], fill=rai_cat_id, outline=rai_cat_id)
                        break
            else:
                for prop in mapping['properties']:
                    if eval(prop):
                        for ann in anns['annotations']:
                            if ann['id'] == obj['id']:
                                objann = ann
                                break
                        rai_cat_id = mapping['properties'][prop]
                        draw.polygon(objann['segmentation'][0], fill=rai_cat_id, outline=rai_cat_id)
                        break

        maskimg.save(os.path.join(outdir, f'{anns["filename"].split(".")[0]}.png'))

    with open(os.path.join(outdir, 'classes.txt'), 'w') as tf:
        for i in range(len(sodcats)):
            if i == 0: continue
            tf.write(f'{i} {sodcats[i]}\n')


"""
UMAP Functions
"""
import os
import requests

def get_umaps(self, umapId=None, datasetId=None, workspaceId=None, cursor=None, limit=None, filters=None, fields=None):
    """Retrieves information about UMAP dataset comparison from the platform.
    
    Parameters
    ----------
    umapId : str
        UMAP Job ID. 
    datasetId : str
        Dataset Id to filter on.
    workspaceId : str
        Workspace ID where the datasets exists.
    cursor : str
        Cursor for pagination.
    limit : int
        Maximum number of umaps to return.
    filters: dict
        Filters that limit output to entries that match the filter 
    fields : list
        List of fields to return, leave empty to get all fields.
    
    Returns
    -------
    dict
        UMAP information.
    """
    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    if limit is not None and limit <= 100: items = limit
    else: items = 100
    umaps = []
    while True:
        if limit and len(umaps) + items > limit: items = limit - len(umaps)
        ret = self.ana_api.getUMAPs(umapId=umapId, datasetId=datasetId, workspaceId=workspaceId, limit=items, cursor=cursor, filters=filters, fields=fields)
        umaps.extend(ret)
        if len(ret) < items or len(umaps) == limit: break
        cursor = ret[-1]["umapId"]
    return umaps
    

def create_umap(self, name, datasetIds, samples, description=None, seed=None, tags=None, workspaceId=None):
    """Creates a UMAP dataset comparison job on the platform.
    
    Parameters
    ----------
    datasetIds : [str]
        Dataset ID to retrieve information for. 
    samples : [int]
        Samples to take from each dataset.
    workspaceId : str
        Workspace ID where the datasets exists.
    
    Returns
    -------
    str
        The UMAP Job ID.
    """
    self.check_logout()
    if name is None: raise ValueError("The name parameter is required!")
    if datasetIds is None: raise ValueError("The datasetIds parameter is required!")
    if samples is None: raise ValueError("The samples parameter is required!")
    if len(datasetIds) != len(samples): raise ValueError("The length of datasetIds must match the length of samples.")
    for sample in samples:
        if sample < 5: raise ValueError("The minimum number of samples is 5!")
    if description is None: description = ''
    if seed is None: seed = 0
    if tags is None: tags = []
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.createUMAP(workspaceId=workspaceId, datasetIds=datasetIds, samples=samples, name=name, description=description, seed=seed, tags=tags)


def delete_umap(self, umapId, workspaceId=None):
    """Deletes/cancels a UMAP dataset comparison on the platform.
    
    Parameters
    ----------
    umapId : str
        UMAP Job ID. 
    workspaceId : str
        Workspace ID where the datasets exists.
    
    Returns
    -------
    bool
        Status.
    """
    self.check_logout()
    if umapId is None: raise ValueError("The umapId parameter is required!")
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.deleteUMAP(umapId=umapId, workspaceId=workspaceId)


def edit_umap(self, umapId, name=None, description=None, tags=None, workspaceId=None):
    """Edits a UMAP dataset comparison on the platform.
    
    Parameters
    ----------
    umapId : str
        UMAP Job ID. 
    name : str
        Name of the UMAP.
    description : str
        Description of the UMAP.
    workspaceId : str
        Workspace ID where the datasets exists.
    
    Returns
    -------
    bool
        Status.
    """
    self.check_logout()
    if umapId is None: raise ValueError("The umapId parameter is required!")
    if name is None: name = ""
    if description is None: description = ""
    if workspaceId is None: workspaceId = self.workspace
    if tags is not None and not isinstance(tags, list): raise ValueError("The tags parameter must be a list!")
    if tags is None: tags = []
    return self.ana_api.editUMAP(umapId=umapId, name=name, description=description, tags=tags, workspaceId=workspaceId)


"""
Datasets API calls.
"""

def getDatasets(self, workspaceId, datasetId=None, cursor=None, limit=100, filters={}, fields=None):
    if fields is None: fields = self.getTypeFields("Dataset")
    fields = "\n".join(fields)
    if filters is None: filters = {}
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getDatasets",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "limit": limit,
                "cursor": cursor,
                "filters": filters
            },
            "query": f"""query 
                getDatasets($workspaceId: String!, $datasetId: String, $cursor: String, $limit: Int, $filters: DatasetFilter) {{
                    getDatasets(workspaceId: $workspaceId, datasetId: $datasetId, cursor: $cursor, limit: $limit, filters: $filters) {{
                       {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getDatasets")


def getDatasetJobs(self, organizationId, workspaceId, datasetId=None, cursor=None, limit=100, filters={}, fields=None):
    if fields is None: fields = self.getTypeFields("DatasetJob")
    fields = "\n".join(fields)
    if filters is None: filters = {}
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getDatasetJobs",
            "variables": {
                "organizationId": organizationId,
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "limit": limit,
                "cursor": cursor,
                "filters": filters
            },
            "query": f"""query 
                getDatasetJobs($organizationId: String, $workspaceId: String, $datasetId: String, $cursor: String, $limit: Int, $filters: DatasetJobFilter) {{
                    getDatasetJobs(organizationId: $organizationId, workspaceId: $workspaceId, datasetId: $datasetId, cursor: $cursor, limit: $limit, filters: $filters) {{
                       {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getDatasetJobs")


def createDataset(self, workspaceId, graphId, name, description=None, runs=1, seed=0, priority=1, tags=[]):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createDataset",
            "variables": {
                "workspaceId": workspaceId,
                "graphId": graphId,
                "name": name,
                "description": description,
                "runs": runs,
                "seed": seed,
                "priority": priority,
                "tags": tags
            },
            "query": """mutation 
                createDataset($workspaceId: String!, $graphId: String!, $name: String!, $description: String, $runs: Int!, $seed: Int!, $priority: Int!, $tags: [String]) {
                    createDataset(workspaceId: $workspaceId, graphId: $graphId, name: $name, description: $description, runs: $runs, seed: $seed, priority: $priority, tags: $tags)
                }"""})
    return self.errorhandler(response, "createDataset")


def deleteDataset(self, workspaceId, datasetId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteDataset",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId
            },
            "query": """mutation 
                deleteDataset($workspaceId: String!, $datasetId: String!) {
                    deleteDataset(workspaceId: $workspaceId, datasetId: $datasetId)
                }"""})
    return self.errorhandler(response, "deleteDataset")


def editDataset(self, workspaceId, datasetId, name=None, description=None, pause=None, priority=None, tags=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editDataset",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "name": name,
                "description": description,
                "pause": pause,
                "priority": priority,
                "tags": tags
            },
            "query": """mutation 
                editDataset($workspaceId: String!, $datasetId: String!, $name: String, $description: String, $pause: Boolean, $priority: Int, $tags: [String]) {
                    editDataset(workspaceId: $workspaceId, datasetId: $datasetId, name: $name, description: $description, pause: $pause, priority: $priority, tags: $tags)
                }"""})
    return self.errorhandler(response, "editDataset")


def downloadDataset(self, workspaceId, datasetId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "downloadDataset",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId
            },
            "query": """mutation 
                downloadDataset($workspaceId: String!, $datasetId: String!) {
                    downloadDataset(workspaceId: $workspaceId, datasetId: $datasetId)
                }"""})
    return self.errorhandler(response, "downloadDataset")


def cancelDataset(self, workspaceId, datasetId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "cancelDataset",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId
            },
            "query": """mutation 
                cancelDataset($workspaceId: String!, $datasetId: String!) {
                    cancelDataset(workspaceId: $workspaceId, datasetId: $datasetId)
                }"""})
    return self.errorhandler(response, "cancelDataset")


def uploadDataset(self, workspaceId, name, filesize, description=None, tags=[]):
    if "UploadDatasetResponse" not in self.fields: fields = self.getTypeFields("UploadDatasetResponse")
    else: fields = self.fields["UploadDatasetResponse"]
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "uploadDataset",
            "variables": {
                "workspaceId": workspaceId,
                "name": name,
                "size": filesize,
                "description": description,
                "tags": tags
            },
            "query": f"""mutation 
                uploadDataset($workspaceId: String!, $name: String!, $size: Float!, $description: String!, $tags: [String]) {{
                    uploadDataset(workspaceId: $workspaceId, name: $name, size: $size, description: $description, tags: $tags){{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "uploadDataset")

def uploadDatasetFinalizer(self, uploadId, key, parts):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "uploadDatasetFinalizer",
            "variables": {
                "uploadId": uploadId,
                "key": key,
                "parts": parts,
            },
            "query": """mutation 
                uploadDatasetFinalizer($uploadId: String!, $key: String!, $parts: [MultipartInput!]) {
                    uploadDatasetFinalizer(uploadId: $uploadId, key: $key, parts: $parts)
                }"""})
    return self.errorhandler(response, "uploadDatasetFinalizer")


def getDatasetRuns(self, workspaceId, datasetId, state=None, fields=None):
    if fields is None: fields = self.getTypeFields("DatasetRun")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getDatasetRuns",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "state": state,
            },
            "query": f"""query 
                getDatasetRuns($workspaceId: String!, $datasetId: String!, $state: String) {{
                    getDatasetRuns(workspaceId: $workspaceId, datasetId: $datasetId, state: $state) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getDatasetRuns")


def getDatasetLog(self, workspaceId, datasetId, runId, fields=None):
    if fields is None: fields = self.getTypeFields("DatasetLog")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getDatasetLog",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "runId": runId
            },
            "query": f"""query 
                getDatasetLog($workspaceId: String!, $datasetId: String!, $runId: String!) {{
                    getDatasetLog(workspaceId: $workspaceId, datasetId: $datasetId, runId: $runId) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getDatasetLog")


def getDatasetFiles(self, workspaceId, datasetId, path, limit=100, cursor=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getDatasetFiles",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "path": path,
                "limit": limit,
                "cursor": cursor
            },
            "query": """query 
                getDatasetFiles($workspaceId: String!, $datasetId: String!, $path: String, $limit: Int, $cursor: String) {
                    getDatasetFiles(workspaceId: $workspaceId, datasetId: $datasetId, path: $path, limit: $limit, cursor: $cursor)
                }"""})
    return self.errorhandler(response, "getDatasetFiles")


def createMixedDataset(self, workspaceId, name, parameters, description='', seed=0, tags=[]):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createMixedDataset",
            "variables": {
                "workspaceId": workspaceId,
                "name": name,
                "parameters": parameters,
                "description": description,
                "seed": seed,
                "tags": tags
            },
            "query": """mutation 
                createMixedDataset($workspaceId: String!, $name: String!, $parameters: String!, $description: String, $seed: Int, $tags: [String]) {
                    createMixedDataset(workspaceId: $workspaceId, name: $name, parameters: $parameters, description: $description, seed: $seed, tags: $tags)
                }"""})
    return self.errorhandler(response, "createMixedDataset")


"""
Editor Functions
"""

import sys
import os
import time
import textwrap
import subprocess
from anatools.lib.print import print_color
from anatools.anaclient._menu import _get_key, _clear_lines, print_link

def _select_session(sessions, action_name="select"):
    """
    Interactive session selector using arrow keys.
    
    Parameters
    ----------
    sessions : list
        List of session dictionaries from listRemoteDevelopment
    action_name : str
        Name of the action (e.g., "stop" or "delete") for display purposes
        
    Returns
    -------
    str or None
        Selected session ID or None if cancelled
    """
    if not sessions:
        print(f"\n❌ No sessions available to {action_name}")
        return None

    current_selection = None
    last_displayed_lines = 0

    def display_sessions():
        nonlocal last_displayed_lines
        if last_displayed_lines > 0:
            _clear_lines(last_displayed_lines)
        
        print(f"\n📝 Use arrow keys (↑/↓) to {action_name} a session, Enter to confirm, q to quit:\n")
        
        for i, session in enumerate(sessions):
            is_selected = current_selection is not None and i == current_selection
            session_line = (
                f"  {'▶' if is_selected else ' '} "
                f"🏢 {session['organization'][:15]} "
                f"🔗 {session['editorUrl']} "
                f"📦 {session['channel']} "
                f"📟 {session['instanceType']} "
                f"📊 {session['status']['state']} "
                f"🔑 {session['sshPort']}"
            )
            if is_selected:
                print(f"\033[44m{session_line}\033[0m")
            else:
                print(session_line)
        
        last_displayed_lines = len(sessions) + 3

    display_sessions()

    while True:
        key = _get_key(None)
        if key == 'up':
            if current_selection is None:
                current_selection = len(sessions) - 1
            else:
                current_selection = max(0, current_selection - 1)
            display_sessions()
        elif key == 'down':
            if current_selection is None:
                current_selection = 0
            else:
                current_selection = min(len(sessions) - 1, current_selection + 1)
            display_sessions()
        elif key == 'enter':
            if current_selection is not None:
                return sessions[current_selection]['editorSessionId']
        elif key.lower() == 'q':
            if last_displayed_lines > 0:
                _clear_lines(last_displayed_lines)
            print("\n❌ Session selection cancelled")
            return None

def _select_editor(client, action_name="select"):
    """Interactive editor selector with rich display.

    Parameters
    ----------
    sessions : list[dict]
        List containing dictionaries describing editor sessions. Must include keys:
        - editorSessionId
        - editorUrl
        - instanceType
        - status (with key 'state')
        - workspaceName (optional) or workspaceId
    action_name : str, default "select"
        The verb used in prompt (e.g. "delete", "stop").

    Returns
    -------
    str | None
        The selected `editorSessionId` or None if user cancelled.
    """
    def _get_sessions(client):
        raw_sessions = client.get_servers()
        workspaces = client.get_workspaces()
        sessions = []
        for s in raw_sessions:
            workspace = [ws for ws in workspaces if ws['workspaceId'] == s.get('workspaceId', '')][0]
            sessions.append({
                'editorSessionId': s.get('editorId'),
                'name': s.get('name', s.get('editorId')),
                'editorUrl': s.get('url', ''),
                'serverPath': s.get('serverPath', '/workspace/'),
                'organization': s.get('organizationId', ''),
                'instanceType': s.get('instanceType', ''),
                'status': s.get('status', {}),
                'workspaceId': s.get('workspaceId', ''),
                'workspaceName': workspace['name']
            })
        return sessions
    
    sessions = _get_sessions(client)
    if not sessions:
        print(f"\n❌ No sessions available to {action_name}")
        return None

    current_selection: int | None = None
    last_lines_printed = 0

    def _link(text: str, url: str) -> str:
        """Return clickable terminal hyperlink (OSC 8)."""
        ESC = "\033"
        return f"{ESC}]8;;{url}{ESC}\\{text}{ESC}]8;;{ESC}\\"


    def _render():
        nonlocal last_lines_printed
        if last_lines_printed > 0: _clear_lines(last_lines_printed)
        print(
            f"\n📝 Use arrow keys (↑/↓) to {action_name} a session, Enter to confirm.\n\n" +
            f"   |       Server      |     Status      |    Instance    |      Workspace    |                IDEs              |\n" +
            f"   |-------------------|-----------------|----------------|-------------------|----------------------------------|"
        )
        for i, s in enumerate(sessions):
            selected = current_selection is not None and i == current_selection
            name = s['name']
            eid = s["editorSessionId"]
            path= s['serverPath']
            host = s["editorUrl"].replace("https://", "")
            if s['status']['state'] == 'running':
                browser = _link("Browser", s["editorUrl"])
                windsurf = _link("Windsurf", f"windsurf://vscode-remote/ssh-remote+{eid}@{host}:443{path}")
                cursor = _link("Cursor", f"cursor://vscode-remote/ssh-remote+{eid}@{host}:443{path}")
                vscode = _link("VSCode", f"vscode://vscode-remote/ssh-remote+{eid}@{host}:443{path}")
            else: browser = windsurf = cursor = vscode = ""
            workspace_name = s.get("workspaceName") or s.get("workspaceId", "")
            line = (
                f"  {'▶' if selected else ' '} "
                f"🆔 {name:<17}  "
                f"📊 {s['status']['state']:<13}  "
                f"📟 {s['instanceType']:<12}  "
                f"🗄 {workspace_name[:15]:<15}  "
                f"{browser}{'  ' if browser else ''}{windsurf}{'  ' if windsurf else ''}{cursor}{'  ' if cursor else ''}{vscode}"
            )
            if selected: print(f"\033[44m{line}\033[0m")
            else: print(line)
        last_lines_printed = len(sessions) + 5

    # First render
    sessions = _get_sessions(client)
    _render()
    while True:
        timeout = None
        for session in sessions:
            if session['status']['state'] in ['starting', 'stopping', 'terminating']: timeout = 1
        key = _get_key(None, timeout)
        if key == 'up':
            if current_selection is None: current_selection = len(sessions) - 1
            else: current_selection = max(0, current_selection - 1)
            _render()
        elif key == 'down':
            if current_selection is None: current_selection = 0
            else: current_selection = min(len(sessions) - 1, current_selection + 1)
            _render()
        elif key == 'enter':
            if current_selection is not None:
                _clear_lines(last_lines_printed)
                return sessions[current_selection]['editorSessionId'], sessions[current_selection]['status']['state']
        elif key == None:
            sessions = _get_sessions(client)
            _render()


def _spinner_animation():
    """Generator for a simple spinner animation."""
    while True:
        for char in '|/-\\':
            yield char

def _show_operation_status(operation):
    """Show a loading spinner while an operation is in progress."""
    import threading
    import time

    stop_thread = threading.Event()
    spinner = _spinner_animation()

    def spin():
        while not stop_thread.is_set():
            sys.stdout.write(f"\r⏳ {next(spinner)} {operation}...")
            sys.stdout.flush()
            time.sleep(0.1)

        sys.stdout.write('\r\033[K')
        sys.stdout.write('\n')
        sys.stdout.flush()

    spinner_thread = threading.Thread(target=spin)
    spinner_thread.start()
    return stop_thread, spinner_thread

def _cleanup_ssh_config(environment, editorSessionId, sshPort=None):
    """Helper function to clean up SSH configuration and known hosts.
    
    Parameters
    ----------
    environment : str
        The environment (dev, prod, etc)
    editorSessionId : str
        The ID of the editor session to clean up.
    sshPort : str, optional
        The SSH port to remove from known hosts. If not provided, only the port-less entry is removed.
    """
    
    if sshPort:
        subprocess.run(["ssh-keygen", "-R", f"[{editorSessionId}.dyn-editor.{environment}.rendered.ai]:{sshPort}"], check=True)

    # Clean up SSH config
    ssh_config_path = os.path.join(os.path.expanduser("~"), ".ssh", "config")
    if os.path.exists(ssh_config_path):
        with open(ssh_config_path, "r") as f:
            ssh_config_content = f.read()
        # Delete the entire block for the old session
        start_pattern = f"Host {editorSessionId}.dyn-editor.{environment}.rendered.ai"
        end_pattern = "Host "
        start_index = ssh_config_content.find(start_pattern)
        if start_index != -1:
            end_index = ssh_config_content.find(end_pattern, start_index + len(start_pattern))
            if end_index == -1:
                end_index = len(ssh_config_content)
            ssh_config_content = ssh_config_content[:start_index] + ssh_config_content[end_index:]
        with open(ssh_config_path, "w") as f:
            f.write(ssh_config_content.strip() + "\n")

def _is_windows_without_wsl():
    """Check if running on Windows without WSL"""
    if sys.platform != "win32":
        return False
    # Check for WSL
    try:
        with open('/proc/version', 'r') as f:
            if 'Microsoft' in f.read():
                return False  # Running in WSL
    except FileNotFoundError:
        pass
    return True  # Regular Windows

def create_remote_development(self, channelId, organizationId=None, channelVersion=None, instanceType=None):
    """
    Creates a remote development environment.

    This method initiates a remote development session on the specified channel, optionally within a given organization.
    If no organizationId is provided, it defaults to the organization associated with the current user.

    Parameters
    ----------
    channelId : str
        The ID of the channel to use for creating the remote development session.
    channelVersion : str, optional
        The version of the channel to use. If not provided, defaults to the latest version.
    organizationId : str, optional
        The ID of the organization where the session will be created. 
        If not provided, defaults to the user's organization.
    instanceType : str, optional
        The type of instance to use for the remote development session.
        If not provided, defaults to the instance type specified in the channel.

    Returns
    -------
    str
        A message indicating that the session is being created, along with a link to access the session.

    Notes
    -----
    - This function checks if the user is logged out before proceeding.
    - Calls `ana_api.createRemoteDevelopment` to initiate the session.
    - Displays a warning message indicating that the feature is experimental.

    Example Output
    --------------
    ⚠️ Warning: This feature is very experimental. Use with caution! ⚠️
    🚀 Your environment will be available here shortly: 🔗 <editorUrl> 🌐
    """
    if self.check_logout():
        print("\n❌ Error: You are not logged in. Please log in first.\n")
        return

    try:
        session = self.ana_api.createRemoteDevelopment(
            channelId=channelId,
            channelVersion=channelVersion,
            instanceType=instanceType
        )

        print(
            "\n⚠️ Warning: This feature is very experimental. Use with caution! ⚠️\n"
            f"🚀 Your environment will be available here shortly: "
            f"🔗 {session['editorUrl']} 🌐\n"
        )
    except Exception as e:
        error_msg = str(e)
        if "403" in error_msg or "Forbidden" in error_msg:
            print("\n❌ Error: Access denied. Please check that:")
            print("  • You have the correct permissions for this channel")
            print("  • The channel ID is correct")
            print("  • You are a member of the organization that owns this channel\n")
        elif "404" in error_msg or "Not Found" in error_msg:
            print("\n❌ Error: Channel not found. Please verify the channel ID is correct.\n")
        else:
            print(f"\n❌ Error: Failed to create remote development environment: {error_msg}\n")
            print("Please make sure you are logged in, the channel ID is valid and you have permission to the channel. If the problem persists, please contact support.\n")

def delete_remote_development(self, editorSessionId=None):
    """
    Deletes a remote development session.

    This method removes a specific editor session, optionally within a given organization.
    If no editorSessionId is provided, it will show a list of active sessions and prompt for selection.

    Parameters
    ----------
    editorSessionId : str, optional
        The ID of the editor session to be deleted. If not provided, will prompt for selection.

    Returns
    -------
    dict
        A dictionary representing the result of the session deletion.

    Notes
    -----
    - This function checks if the user is logged out before proceeding.
    - Calls `ana_api.deleteRemoteDevelopment` to perform the deletion.
    - Use arrow keys (↑/↓) to select a session, Enter to confirm, q to quit
    """
    if self.check_logout():
        print("\n❌ Error: You are not logged in. Please log in first.\n")
        return

    if editorSessionId is None:
        sessions = self.ana_api.listRemoteDevelopment(organizationId=None)
        editorSessionId = _select_session(sessions, action_name="delete")
        if editorSessionId is None:
            return

    spinner_stop_event, spinner_thread = _show_operation_status(f"Deleting Development Session {editorSessionId}")
    try:
        session = self.ana_api.deleteRemoteDevelopment(editorSessionId=editorSessionId)
    except Exception as e:
        error_msg = str(e)
        if "403" in error_msg or "Forbidden" in error_msg:
            print("\n❌ Error: Access denied. Please check that:")
            print("  • You have the correct permissions for this channel")
            print("  • The channel ID is correct")
            print("  • You are a member of the organization that owns this channel\n")
        elif "404" in error_msg or "Not Found" in error_msg:
            print("\n❌ Error: Session not found. Please verify the session ID is correct.\n")
        else:
            print(f"\n❌ Error: Failed to delete development session: {error_msg}\n")
            print("If the problem persists, please contact support.\n")
    finally:
        spinner_stop_event.set()
        spinner_thread.join()
    print(f"\n🗑️  Successfully deleted Development Session {editorSessionId}\n")

    print(f"To remove SSH configuration call `remove_ssh_remote_development()`")


def list_remote_development(self, organizationId=None): 
    """Shows all the active development sessions in the organization.
    
    Parameters
    ----------
    organizationId : str
        The ID of the organization to list the active development sessions.
    
    Returns
    -------
    list[dict]
        If organizationId is not provided, returns all active sessions in organizations that user has access to.
        If organizationId is provided, returns active sessions in that specific organization.
    """
    if self.check_logout():
        print("\n❌ Error: You are not logged in. Please log in first.\n")
        return

    try:
        sessions = self.ana_api.listRemoteDevelopment(organizationId=organizationId)

        if not sessions:
            print("✨ No active development sessions found. Use `create_remote_development` to start a new session.")
            return sessions

        if organizationId is None:
            print("\n🚧 Active Development Sessions:\n")
        else:
            print(f"\n🚧 Active Development Sessions in Organization {organizationId}:\n")
        

        for session in sessions:
            print(
                f"🏢 {session['organization'][:15]} "
                f"🔗 {session['editorUrl']} "
                f"📦 {session['channel']} "
                f"📟 {session['instanceType']} "
                f"📊 {session['status']['state']} "
                f"🔑 {session['sshPort']}"
                f"📅 {session['updatedAt']}"
            )
            if session['status']['state'] == 'ERROR':
                print(f"\t❌ Error: {session['status']['message']}")

        print(f"\n\nTo ssh in to an instance, add your public ssh key to /home/anadev/.ssh/authorized_keys and call `prepare_ssh_remote_development()`.")
    except Exception as e:
        error_msg = str(e)
        if "403" in error_msg or "Forbidden" in error_msg:
            print("\n❌ Error: Access denied. Please check that:")
            print("  • You have the correct permissions for this channel")
            print("  • The channel ID is correct")
            print("  • You are a member of the organization that owns this channel\n")
        elif "404" in error_msg or "Not Found" in error_msg:
            print("\n❌ Error: Organization not found. Please verify the organization ID is correct.\n")
        else:
            print(f"\n❌ Error: Failed to list development sessions: {error_msg}\n")
            print("If the problem persists, please contact support.\n")


def stop_remote_development(self, editorSessionId=None):
    """
    Stops a remote development session.

    This method stops a specific editor session, optionally within a given organization.
    If no editorSessionId is provided, it will show a list of active sessions and prompt for selection.

    Parameters
    ----------
    editorSessionId : str, optional
        The ID of the editor session to be stopped. If not provided, will prompt for selection.

    Returns
    -------
    dict
        A dictionary representing the result of the session stop operation.

    Notes
    -----
    - This function checks if the user is logged out before proceeding.
    - Calls `ana_api.stopRemoteDevelopment` to stop the session.
    - Use arrow keys (↑/↓) to select a session, Enter to confirm, q to quit
    - Only shows sessions that are currently running or resuming
    """
    if self.check_logout():
        print("\n❌ Error: You are not logged in. Please log in first.\n")
        return

    if editorSessionId is None:
        sessions = self.ana_api.listRemoteDevelopment(organizationId=None)
        active_sessions = [s for s in sessions if s['status']['state'] in ('RUNNING', 'RESUMING')]
        if not active_sessions and sessions:
            print("✨ No active sessions available to stop.")
            return
        editorSessionId = _select_session(active_sessions, action_name="stop")
        if editorSessionId is None:
            return

    spinner_stop_event, spinner_thread = _show_operation_status(f"Stopping Development Session {editorSessionId}")
    try:
        session = self.ana_api.stopRemoteDevelopment(editorSessionId=editorSessionId)
    except Exception as e:
        error_msg = str(e)
        if "403" in error_msg or "Forbidden" in error_msg:
            print("\n❌ Error: Access denied. Please check that:")
            print("  • You have the correct permissions for this channel")
            print("  • The channel ID is correct")
            print("  • You are a member of the organization that owns this channel\n")
        elif "404" in error_msg or "Not Found" in error_msg:
            print("\n❌ Error: Session not found. Please verify the session ID is correct.\n")
        else:
            print(f"\n❌ Error: Failed to stop development session: {error_msg}\n")
            print("If the problem persists, please contact support.\n")
    finally:
        spinner_stop_event.set()
        spinner_thread.join()
    print(f"\n🛑 Successfully stopped Development Session {editorSessionId}\n")


def start_remote_development(self, editorSessionId=None):
    """
    Starts a remote development session.

    This method starts a specific editor session, optionally within a given organization.
    If no editorSessionId is provided, it will show a list of stopped sessions and prompt for selection.

    Parameters
    ----------
    editorSessionId : str, optional
        The ID of the editor session to be started. If not provided, will prompt for selection.

    Returns
    -------
    dict
        A dictionary representing the result of the session start operation.

    Notes
    -----
    - This function checks if the user is logged out before proceeding.
    - Calls `ana_api.startRemoteDevelopment` to start the session.
    - Use arrow keys (↑/↓) to select a session, Enter to confirm, q to quit
    - Only shows sessions that are currently stopped
    """
    if self.check_logout():
        print("\n❌ Error: You are not logged in. Please log in first.\n")
        return

    if editorSessionId is None:
        sessions = self.ana_api.listRemoteDevelopment(organizationId=None)
        stopped_sessions = [s for s in sessions if s['status']['state'] == 'STOPPED']
        if not stopped_sessions and sessions:
            print("✨ No stopped sessions available to start.")
            return
        editorSessionId = _select_session(stopped_sessions, action_name="start")
        if editorSessionId is None:
            return

    spinner_stop_event, spinner_thread = _show_operation_status(f"Starting Development Session {editorSessionId}")
    try:
        session = self.ana_api.startRemoteDevelopment(editorSessionId=editorSessionId)
    except Exception as e:
        error_msg = str(e)
        if "403" in error_msg or "Forbidden" in error_msg:
            print("\n❌ Error: Access denied. Please check that:")
            print("  • You have the correct permissions for this channel")
            print("  • The channel ID is correct")
            print("  • You are a member of the organization that owns this channel\n")
        elif "404" in error_msg or "Not Found" in error_msg:
            print("\n❌ Error: Session not found. Please verify the session ID is correct.\n")
        else:
            print(f"\n❌ Error: Failed to start development session: {error_msg}\n")
            print("If the problem persists, please contact support.\n")
    finally:
        spinner_stop_event.set()
        spinner_thread.join()
    print(
        f"\n🚀 Successfully started Development Session {editorSessionId}\n"
        f"🔗 Your environment will be available here shortly: {session['editorUrl']} 🌐\n"
    )

# create a new function to prepare ssh sessions. clear old sessions from hosts file. we should list sessions like we do above and allow users to use keys to select the environment
def prepare_ssh_remote_development(self, editorSessionId=None, forceUpdate=False):
    """
    Prepares a remote development session for SSH access.

    This method prepares a specific editor session for SSH access, optionally within a given organization.
    If no editorSessionId is provided, it will show a list of active sessions and prompt for selection.

    Parameters
    ----------
    editorSessionId : str, optional
        The ID of the editor session to prepare SSH for. If not provided, will prompt user to select.
    forceUpdate : bool, optional
        If True, will remove existing SSH configuration before adding new one.

    Returns
    -------
    dict
        A dictionary representing the result of the session preparation.

    Notes
    -----
    - This function checks if the user is logged out before proceeding.
    - Use arrow keys (↑/↓) to select a session, Enter to confirm, q to quit
    - Only shows sessions that are currently running or resuming
    """
    if self.check_logout():
        print("\n❌ Error: You are not logged in. Please log in first.\n")
        return

    if _is_windows_without_wsl():
        print("\n⚠️  SSH configuration management is not supported on Windows.")
        print("Please use Windows Subsystem for Linux (WSL) for full SSH support.")
        return

    sessions = self.ana_api.listRemoteDevelopment(organizationId=None)
    active_sessions = [s for s in sessions if s['status']['state'] in ('RUNNING', 'RESUMING')]
    if not active_sessions and sessions:
        print("✨ No active sessions available to prepare.")
        return

    if editorSessionId is None:
        editorSessionId = _select_session(active_sessions, action_name="prepare")
        if editorSessionId is None:
            return
    
    # Find the selected session to get its port
    session = next((s for s in active_sessions if s['editorSessionId'] == editorSessionId), None)
    if not session:
        print(f"❌ Could not find session {editorSessionId}")
        return

    spinner_stop_event, spinner_thread = _show_operation_status(f"Preparing SSH for Development Session {editorSessionId}")
    try:
        if forceUpdate:
            _cleanup_ssh_config(self.environment, editorSessionId, session.get('sshPort'))
        
        ssh_config_path = os.path.join(os.path.expanduser("~"), ".ssh", "config")
        with open(ssh_config_path, "a+") as f:
            f.seek(0)
            if f.read().find(f"Host {editorSessionId}.dyn-editor.{self.environment}.rendered.ai") == -1:
                ssh_config_template = f'''
                    Host {editorSessionId}.dyn-editor.{self.environment}.rendered.ai
                        HostName ssh-editor.{self.environment}.rendered.ai
                        Port {session.get('sshPort')}
                        User anadev
                        StrictHostKeyChecking no
                        UserKnownHostsFile /dev/null
                '''
                f.write(textwrap.dedent(ssh_config_template))
    except Exception as e:
        error_msg = str(e)
        print(f"\n❌ Error: Failed to prepare SSH for development session: {error_msg}\n")
        print("If the problem persists, please contact support.\n")
    finally:
        spinner_stop_event.set()
        spinner_thread.join()
    print(f"\n🚀 Successfully prepared SSH for Development Session {editorSessionId}\n")

    print(f"To SSH in to an instance, add your public ssh key to /home/anadev/.ssh/authorized_keys")
    print(f"then call `ssh {editorSessionId}.dyn-editor.{self.environment}.rendered.ai`")
    print(f"or use host {editorSessionId}.dyn-editor.{self.environment}.rendered.ai in your code editor\n")

    print(f"To remove SSH configuration call `remove_ssh_remote_development()`\n")

# add a new function to remove ssh sessions from hosts file and clear old sessions from config
def remove_ssh_remote_development(self, editorSessionId=None):
    """
    Removes a remote development session from SSH access.

    This method removes a specific editor session from SSH access, optionally within a given organization.
    If no editorSessionId is provided, it will show a list of active sessions and prompt for selection.

    Parameters
    ----------
    editorSessionId : str, optional
        The ID of the editor session to be removed. If not provided, will prompt user to select.

    Returns
    -------
    dict
        A dictionary representing the result of the session removal.

    Notes
    -----
    - This function checks if the user is logged out before proceeding.
    - Use arrow keys (↑/↓) to select a session, Enter to confirm, q to quit
    - Only shows sessions that are currently running or resuming
    """
    if self.check_logout():
        print("\n❌ Error: You are not logged in. Please log in first.\n")
        return

    if _is_windows_without_wsl():
        print("\n⚠️  SSH configuration management is not supported on Windows.")
        print("Please use Windows Subsystem for Linux (WSL) for full SSH support.")
        return

    sessions = self.ana_api.listRemoteDevelopment(organizationId=None)

    if editorSessionId is None:
        editorSessionId = _select_session(sessions, action_name="remove")
        if editorSessionId is None:
            return
    
    # Find the selected session to get its port
    session = next((s for s in sessions if s['editorSessionId'] == editorSessionId), None)
    if not session:
        print(f"❌ Could not find session {editorSessionId}")
        return

    spinner_stop_event, spinner_thread = _show_operation_status(f"Removing SSH for Development Session {editorSessionId}")
    try:
        _cleanup_ssh_config(self.environment, editorSessionId, session.get('sshPort'))
    except Exception as e:
        error_msg = str(e)
        print(f"\n❌ Error: Failed to remove SSH for development session: {error_msg}\n")
        print("If the problem persists, please contact support.\n")
    finally:
        spinner_stop_event.set()
        spinner_thread.join()
    print(f"\n🛑 Successfully removed SSH for Development Session {editorSessionId}\n")


def invite_remote_development(self, editorSessionId, email):
    """
    Invites a user to join a remote development session.

    Parameters
    ----------
    editorSessionId : str
        The ID of the editor session to invite the user to.
    email : str
        The email address of the user to invite.
    Returns
    -------
    bool
        A boolean status of whether the operation was successful.
    """
    self.check_logout()
    return self.ana_api.inviteRemoteDevelopment(editorSessionId, email)


def register_ssh_key(self, filename=None):
    """
    Registers a public SSH key for use with remote development sessions.

    Parameters
    ----------
    filename : str
        The filename of the .pub SSH key to register.
    Returns
    -------
    bool
        A boolean status of whether the operation was successful.
    """
    if self.check_logout():
        print("\n❌ Error: You are not logged in. Please log in first.\n")
        return

    if filename is None:
        # Look for SSH keys in standard locations
        ssh_dir = os.path.expanduser("~/.ssh")
        if not os.path.exists(ssh_dir):
            print("\n❌ Error: No SSH directory found at ~/.ssh\n")
            return
        
        # Find all .pub files
        pub_keys = []
        for file in os.listdir(ssh_dir):
            if file.endswith(".pub"):
                pub_keys.append(os.path.join(ssh_dir, file))
        
        if not pub_keys:
            print("\n❌ Error: No public SSH keys found in ~/.ssh\n")
            return
        
        # Present selection interface
        print("\nAvailable SSH public keys:\n")
        for i, key in enumerate(pub_keys, 1):
            print(f"{i}. {os.path.basename(key)}")
        
        while True:
            try:
                choice = input("\nSelect a key number (or 'q' to quit): ")
                if choice.lower() == 'q':
                    return
                choice = int(choice)
                if 1 <= choice <= len(pub_keys):
                    filename = pub_keys[choice - 1]
                    break
                print("Invalid selection. Please try again.")
            except ValueError:
                print("Invalid input. Please enter a number or 'q' to quit.")
    
    # Validate and process the selected/provided key
    if not os.path.splitext(filename)[1] == ".pub":
        raise Exception(f"Invalid filename. Please provide a public SSH key that ends with .pub.")
    
    name = os.path.splitext(os.path.basename(filename))[0]
    if os.path.isfile(filename):
        with open(filename, "r") as f:
            publicKey = f.read().strip()
    else:
        raise Exception(f"Could not find SSH public key file {filename}.")
    
    return self.ana_api.createSSHKey(name=name, key=publicKey)


def deregister_ssh_key(self, name=None):
    """
    Removes a public SSH key for use with remote development sessions.

    Parameters
    ----------
    name : str
        The name of the SSH key to deregister.
    Returns
    -------
    bool
        A boolean status of whether the operation was successful.
    """
    if self.check_logout():
        print("\n❌ Error: You are not logged in. Please log in first.\n")
        return

    if name is None:
        # Get list of registered SSH keys
        keys = self.ana_api.getSSHKeys()
        if not keys:
            print("\n❌ No SSH keys registered\n")
            return

        # Present selection interface
        print("\nRegistered SSH keys:\n")
        for i, key in enumerate(keys, 1):
            print(f"{i}. {key['name']}")

        while True:
            try:
                choice = input("\nSelect a key number to remove (or 'q' to quit): ")
                if choice.lower() == 'q':
                    return
                choice = int(choice)
                if 1 <= choice <= len(keys):
                    name = keys[choice - 1]['name']
                    break
                print("Invalid selection. Please try again.")
            except ValueError:
                print("Invalid input. Please enter a number or 'q' to quit.")

    result = self.ana_api.deleteSSHKey(name=name)
    if result:
        print(f"\n✅ Successfully removed SSH key: {name}\n")
    return result


def get_ssh_keys(self):
    """
    Returns a list of SSH keys a user has registered with the platform.

    Parameters
    ----------
    Returns
    -------
    list
        A list of registered SSH keys.
    """
    if self.check_logout():
        print("\n❌ Error: You are not logged in. Please log in first.\n")
        return
    return self.ana_api.getSSHKeys()


# Service/Workspace Editors
def get_servers(self, organizationId=None, workspaceId=None, serverId=None, cursor=None, limit=None, filters=None, fields=None):
    """
    Returns a list of servers for a workspace.
    
    Parameters
    ----------
    organizationId : str
        The ID of the organization to get servers for.
    workspaceid : str
        The ID of the workspace to get servers for.
    serverId : str
        The ID of the server to get.
    Returns
    -------
    list
        A list of servers for the specified workspace.
    """
    self.check_logout()
    if limit is None: limit = 100
    editors = self.ana_api.getEditors(organizationId=organizationId, workspaceId=workspaceId, editorId=serverId, cursor=cursor, limit=limit, filters=filters, fields=fields)
    if not self.interactive or len(editors) != 1: return editors
    else:
        editorId = editors[0]['editorId']
        while True:
            editor = self.ana_api.getEditors(editorId=editorId)[0]
            print(f'Waiting for server {editorId} to start...', end='\r')
            if editor['status']['state'] == 'running': break
            time.sleep(5)
        print_color(f'✅ Server available, open with:                ', '91e600')
        serverpath = editor.get('serverPath', '/workspace/')
        print("  ", end=''); print_link("Browser", editor["url"])
        print("  ", end=''); print_link("Cursor", f"cursor://vscode-remote/ssh-remote+{editor['editorId']}@{editor['url'].replace('https://', '')}:443{serverpath}")
        print("  ", end=''); print_link("VSCode", f"vscode://vscode-remote/ssh-remote+{editor['editorId']}@{editor['url'].replace('https://', '')}:443{serverpath}")
        print("  ", end=''); print_link("Windsurf", f"windsurf://vscode-remote/ssh-remote+{editor['editorId']}@{editor['url'].replace('https://', '')}:443{serverpath}")
        return

def create_server(self, organizationId=None, workspaceId=None, instance=None, name=None):
    """
    Creates a new server for a workspace.

    Parameters
    ----------
    organizationId : str
        The ID of the organization to create the server for.
    workspaceId : str
        The ID of the workspace to create the server for.
    instance : str
        The instance of the server to create.
    name : str
        A custom name to give the server.
    Returns
    -------
    str
        The ID for the newly created server.
    """
    import time
    self.check_logout()
    if organizationId is None: organizationId = self.organization
    if workspaceId is None: workspaceId = self.workspace
    editorId = self.ana_api.createEditor(organizationId=organizationId, workspaceId=workspaceId, instance=instance, name=name)
    if not self.interactive: return editorId
    else:
        starttime = time.time()
        while True:
            editor = self.ana_api.getEditors(editorId=editorId)[0]
            print(f'Waiting for server {editorId} to start...[{time.time()-starttime:.3f}s]', end='\r')
            if editor['status']['state'] == 'running':
                break
            time.sleep(2)
        print(f'Waiting for server {editorId} to start...done. [{time.time()-starttime:3f}s]')
        print_color(f'✅ Server available, open with:                ', '91e600')
        serverpath = editor.get('serverPath', '/workspace/')
        print("  ", end=''); print_link("Browser", editor["url"])
        print("  ", end=''); print_link("Cursor", f"cursor://vscode-remote/ssh-remote+{editor['editorId']}@{editor['url'].replace('https://', '')}:443{serverpath}")
        print("  ", end=''); print_link("VSCode", f"vscode://vscode-remote/ssh-remote+{editor['editorId']}@{editor['url'].replace('https://', '')}:443{serverpath}")
        print("  ", end=''); print_link("Windsurf", f"windsurf://vscode-remote/ssh-remote+{editor['editorId']}@{editor['url'].replace('https://', '')}:443{serverpath}")
        return


def delete_server(self, serverId):
    """
    Deletes an editor session.
    
    Parameters
    ----------
    serverId : str
        The ID of the server to delete.
    Returns
    -------
    bool
        If true, the editor session was successfully deleted.
    """
    self.check_logout()
    return self.ana_api.deleteEditor(editorId=serverId)

    
def edit_server(self, serverId, name):
    """
    Edit the name of a server.
    
    Parameters
    ----------
    serverId : str
        The ID of the editor session to invite the user to.
    name : str
        The name to give the server.
    Returns
    -------
    bool
        If true, the server was successfully edited.
    """
    self.check_logout()
    return self.ana_api.editEditor(editorId=serverId, name=name)


def start_server(self, serverId):
    """
    Starts a server.
    
    Parameters
    ----------
    serverId : str
        The ID of the server to start.
    Returns
    -------
    bool
        If true, the server was successfully started.
    """
    import time
    self.check_logout()
    status =  self.ana_api.startEditor(editorId=serverId)
    if not self.interactive: return status
    else:
        starttime = time.time()
        while True:
            resp = self.ana_api.getEditors(editorId=serverId)
            if not resp:
                time.sleep(5)
                continue
            editor = resp[0]
            print(f'Waiting for editor {serverId} to start...[{time.time()-starttime:.3f}s]', end='\r')
            if editor['status']['state'] == 'running':
                break
            time.sleep(2)
        print(f'Waiting for editor {serverId} to start...done. [{time.time()-starttime:3f}s]')
        print_color(f'✅ Editor available, open with:                ', '91e600')
        serverpath = editor.get('serverPath', '/workspace/')
        print("  ", end=''); print_link("Browser", editor["url"])
        print("  ", end=''); print_link("Cursor", f"cursor://vscode-remote/ssh-remote+{editor['editorId']}@{editor['url'].replace('https://', '')}:443{serverpath}")
        print("  ", end=''); print_link("VSCode", f"vscode://vscode-remote/ssh-remote+{editor['editorId']}@{editor['url'].replace('https://', '')}:443{serverpath}")
        print("  ", end=''); print_link("Windsurf", f"windsurf://vscode-remote/ssh-remote+{editor['editorId']}@{editor['url'].replace('https://', '')}:443{serverpath}")
        return
    

def stop_server(self, serverId):
    """
    Stops a server.
    
    Parameters
    ----------
    editorId : str
        The ID of the server to stop.
    Returns
    -------
    bool
        If true, the server was successfully stopped.
    """
    self.check_logout()
    return self.ana_api.stopEditor(editorId=serverId)
    


# Copyright 2019-2023 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import yaml
import os
import logging

logger = logging.getLogger(__name__)

def file_metadata(filename):
    """
    Get the metadata associated with a file. Return it as a dictionary.
    If there is no metadata file then return an empty dictionary.
    """
    metadata_filename = filename + ".anameta"
    if os.path.isfile(metadata_filename):
        try:
            with open(metadata_filename, "r") as f:
                metadata = yaml.safe_load(f)
        except:
            logger.error(f"Error reading metadata file '{metadata_filename}'")
            raise
    else:
        # if there is no .anameta file then return None
        metadata = None
    return metadata


"""
Organizations API calls.
"""

def getOrganizations(self, organizationId=None, limit=100, cursor=None, fields=None):
    if fields is None: fields = self.getTypeFields("Organization")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getOrganizations",
            "variables": {
                "organizationId": organizationId,
                "limit": limit,
                "cursor": cursor
            },
            "query": f"""query 
                getOrganizations($organizationId: String, $cursor: String, $limit: Int,) {{
                    getOrganizations(organizationId: $organizationId, cursor: $cursor, limit: $limit) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getOrganizations")


def editOrganization(self, organizationId, name):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editOrganization",
            "variables": {
                "organizationId": organizationId,
                "name": name
            },
            "query": """mutation 
                editOrganization($organizationId: String!, $name: String!) {
                    editOrganization(organizationId: $organizationId, name: $name) 
                }"""})
    return self.errorhandler(response, "editOrganization")



import json
import os
from PIL import Image

def convert(size, box):
    """
    Convert bounding box to yolo format:
    https://stackoverflow.com/questions/56115874/how-to-convert-bounding-box-x1-y1-x2-y2-to-yolo-style-x-y-w-h
    """
    dw = 1./size[0]
    dh = 1./size[1]
    x = (box[0] + box[1])/2.0
    y = (box[2] + box[3])/2.0
    w = box[1] - box[0]
    h = box[3] - box[2]
    x = x*dw
    w = w*dw
    y = y*dh
    h = h*dh
    return (x,y,w,h)

def convert_yolo(datadir, outdir, mapping=None):
    """ Generate annotations in YOLO format. Result will be placed in outdir.

    Parameters
    ----------
    datadir : str
        Location of Rendered.ai dataset output.
    outdir : str
        Location where the results should be written.
    mapfile: str
        The map file used for annotations (YAML only).
    
    Returns
    -------
    """
    annsdir = os.path.join(datadir, "annotations")
    metadir = os.path.join(datadir, "metadata")
    imagedir = os.path.join(datadir, "images")
    annsfiles = os.listdir(annsdir)

    # for each interpretation, gather annotations and map categories
    classes = []
    for f in sorted(annsfiles):
        with open(os.path.join(annsdir,f), 'r') as af:
            anns = json.load(af)
        with open(os.path.join(metadir,f.replace('ana','metadata')), 'r') as mf:
            metadata = json.load(mf)
        
        try:
            image = Image.open(os.path.join(imagedir, anns['filename']))
            width = image.size[0]
            height = image.size[1]
        except:
            raise Exception(f'Could not find a supported image for {anns["filename"]} in the dataset images directory.')

        # for each object in the metadata file, check if any of the properties are true
        yolodata = ""
        for obj in metadata['objects']:
            if mapping is None:
                for ann in anns['annotations']:
                    if ann['id'] == obj['id']:
                        size = (width, height)
                        xmin = ann['bbox'][0]
                        ymin = ann['bbox'][1]
                        xmax = ann['bbox'][0] + ann['bbox'][2]
                        ymax = ann['bbox'][1] + ann['bbox'][3]
                        box = (xmin, xmax, ymin, ymax)
                        x, y, w, h = convert(size, box)
                        if obj['type'] not in classes: classes.append(obj['type'])
                        num = classes.index(obj['type'])
                        objectdata = f"{num} {x} {y} {w} {h}\n"
                        yolodata += objectdata
                        break
            else:
                for prop in mapping['properties']:
                    if eval(prop):
                        for ann in anns['annotations']:
                            if ann['id'] == obj['id']:
                                objann = ann
                        size = (width, height)
                        xmin = objann['bbox'][0]
                        ymin = objann['bbox'][1]
                        xmax = objann['bbox'][0] + objann['bbox'][2]
                        ymax = objann['bbox'][1] + objann['bbox'][3]
                        box = (xmin, xmax, ymin, ymax)
                        x, y, w, h = convert(size, box)
                        num = mapping['properties'][prop]
                        objectdata = f"{num} {x} {y} {w} {h}\n"
                        yolodata += objectdata
                        if mapping['classes'][num][-1] not in classes: classes.append(mapping['classes'][num][-1])
                        break
        
        # write xmlfile
        with open(os.path.join(outdir,f.replace('-ana.json','.txt')), 'w+') as txtfile:
            txtfile.write(yolodata)
    
    with open(os.path.join(outdir,'classes.txt'), 'w+') as txtfile:
        for name in classes:
            txtfile.write(f"{name}\n")


import subprocess
import subprocess
import logging
import os
import json
import anatools

def configure_logging():
    """Configures logging to output to both file and console."""
    log_directory = os.path.join(os.path.dirname(__file__), '.logs')
    os.makedirs(log_directory, exist_ok=True)
    log_filepath = os.path.join(log_directory, 'mcp-server.log')

    # Create a custom logger
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    # Create handlers
    c_handler = logging.StreamHandler()
    f_handler = logging.FileHandler(log_filepath)
    c_handler.setLevel(logging.INFO)
    f_handler.setLevel(logging.INFO)

    # Create formatters and add it to handlers
    log_format = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    c_handler.setFormatter(log_format)
    f_handler.setFormatter(log_format)

    # Add handlers to the logger
    # Check if handlers are already added to avoid duplication
    if not logger.handlers:
        logger.addHandler(c_handler)
        logger.addHandler(f_handler)

    return logger

# Configure and get the logger instance
logger = configure_logging()

_anatools_client = None

def get_anatools_client():
    """Initializes and returns a singleton anatools client."""
    global _anatools_client
    if _anatools_client is None:
        _anatools_client = anatools.client()
    return _anatools_client



def get_workspace_id():
    """Gets the Workspace ID from the environment."""
    return os.getenv("RENDEREDAI_WORKSPACE_ID")


async def run_service(service_id: str, params: dict) -> str:
    """Runs a local service container using Docker.

    Args:
        service_id: The ID of the service to run, used to find the Docker image.
        params: A dictionary of parameters to pass to the service as a JSON string payload. The /ana/data directory is mounted to /data in the container.

    Returns:
        A markdown-formatted string with the execution result or an error.
    """
    renderedai_api_key = os.getenv("RENDEREDAI_API_KEY")
    if not renderedai_api_key:
        return """❌ **Configuration Error**: RENDEREDAI_API_KEY is not set.

**Suggestions**:
- Please set the `RENDEREDAI_API_KEY` environment variable in your `.env` file.
"""

    logger.info(f"Attempting to run service: {service_id}")

    try:
        # Find the docker image for the given service_id
        image_name = None
        # Command to list all images in JSON format
        image_list_cmd = "docker images -a --format '{{json .}}'"
        result = subprocess.run(image_list_cmd, shell=True, capture_output=True, text=True, check=True)

        # Find the image that contains the service_id in its JSON representation
        for image_json_str in result.stdout.strip().split('\n'):
            if service_id in image_json_str:
                image_data = json.loads(image_json_str)
                image_name = image_data.get('ID') # Using Image ID is more reliable
                if image_name:
                    break

        if not image_name:
            logger.error(f"No Docker image found for service_id: {service_id}")
            return f"""❌ **Execution Error**: Docker image not found.

**Suggestions**:
- Verify that an image containing the service ID `{service_id}` exists on your local machine.
- Check the output of `docker images`.
"""

        logger.info(f"Found image '{image_name}' for service '{service_id}'.")

        # get user id and group id
        user_id = os.getuid()
        group_id = os.getgid()

        def _map_paths_to_container(obj):
            if isinstance(obj, dict):
                return {k: _map_paths_to_container(v) for k, v in obj.items()}
            elif isinstance(obj, list):
                return [_map_paths_to_container(elem) for elem in obj]
            elif isinstance(obj, str) and '/ana/data' in obj:
                return obj.replace('/ana/data', '/data')
            return obj

        mapped_params = _map_paths_to_container(params)

        logger.info(f"Mapped parameters: {mapped_params}")

        # Safely create the JSON payload string
        payload = json.dumps(mapped_params)

        logger.info(f"Payload: {payload}")

        # Construct and execute the docker command
        docker_command = [
            "docker", "run", "--rm",
            "-u", f"{user_id}:{group_id}",
            "-e", f"payload={payload}",
            "-v", f"{os.path.expanduser('~')}/.renderedai:/data",
            image_name
        ]

        logger.debug(f"Executing command: {' '.join(docker_command)}")

        run_result = subprocess.run(docker_command, capture_output=True, text=True)

        if run_result.returncode != 0:
            logger.error(f"Docker execution failed for {service_id}. Stderr: {run_result.stderr.strip()}")
            return f"""❌ **Docker Execution Error** (Exit Code: {run_result.returncode}):

        **Stderr**:
        ```
        {run_result.stderr.strip()}
        ```

        **Stdout**:
        ```
        {run_result.stdout.strip()}
        ```
        """

        logger.info(f"Service {service_id} executed successfully.")
        return f"""✅ **Service Execution Successful**

        **Service ID**: `{service_id}`

        **Output**:
        ```
        {run_result.stdout.strip()}
        ```
        """

    except subprocess.CalledProcessError as e:
        logger.error(f"Subprocess failed: {e.stderr}")
        return f"""❌ **Subprocess Error**: Failed to execute a system command.

        **Error**:
        ```
        {e.stderr}
        ```
        """
    except Exception as e:
        logger.exception(f"An unexpected error occurred while running service {service_id}")
        return f"""❌ **Unexpected Error**: {e}

        **Suggestions**:
        - Check the server logs for more details.
        """


"""
Services API calls.
"""

def getServiceTypes(self, fields=None):
    if fields is None: fields = self.getTypeFields("ServiceType")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getServiceTypes",
            "variables": {},
            "query": F"""query 
                getServiceTypes {{
                    getServiceTypes {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getServiceTypes")


def getServices(self, organizationId=None, workspaceId=None, serviceId=None, limit=100, cursor=None, filters={}, fields=None):
    if fields is None: fields = self.getTypeFields("Service")
    fields = "\n".join(fields)
    if filters is None: filters = {}
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getServices",
            "variables": {
                "organizationId": organizationId,
                "workspaceId": workspaceId,
                "serviceId": serviceId,
                "limit": limit,
                "cursor": cursor,
                "filters": filters
            },
            "query": F"""query 
                getServices($organizationId: String, $workspaceId: String, $serviceId: String, $limit: Int, $cursor: String, $filters: ServicesFilter) {{
                    getServices(organizationId: $organizationId, workspaceId: $workspaceId, serviceId: $serviceId, limit: $limit, cursor: $cursor, filters: $filters) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getServices")


def createService(self, organizationId, serviceTypeId, name, description=None, volumes=None, instance=None, tags=None, editorId=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createService",
            "variables": {
                "organizationId": organizationId,
                "serviceTypeId": serviceTypeId,
                "name": name,
                "description": description,
                "volumes": volumes,
                "instance": instance,
                "tags": tags,
                "editorId": editorId
            },
            "query": """mutation 
                createService($organizationId: String!, $serviceTypeId: String!, $name: String!, $description: String, $volumes: [String], $instance: String, $tags: [String], $editorId: String) {
                    createService(organizationId: $organizationId, serviceTypeId: $serviceTypeId, name: $name, description: $description, volumes: $volumes, instance: $instance, tags: $tags, editorId: $editorId)
                }"""})
    return self.errorhandler(response, "createService")


def deleteService(self, serviceId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteService",
            "variables": {
                "serviceId": serviceId,
            },
            "query": """mutation 
                deleteService($serviceId: String!) {
                    deleteService(serviceId: $serviceId)
                }"""})
    return self.errorhandler(response, "deleteService")


def editService(self, serviceId, name=None, description=None, volumes=None, instance=None, tags=None, schema=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editService",
            "variables": {
                "serviceId": serviceId,
                "name": name,
                "description": description,
                "volumes": volumes,
                "instance": instance,
                "tags": tags,
                "schema": schema
            },
            "query": """mutation 
                editService($serviceId: String!, $name: String, $description: String, $volumes: [String], $instance: String, $tags: [String], $schema: String) {
                    editService(serviceId: $serviceId, name: $name, description: $description, volumes: $volumes, instance: $instance, tags: $tags, schema: $schema)
                }"""})
    return self.errorhandler(response, "editService")


def deployService(self, serviceId):
    fields = self.getTypeFields("ECRDeployment")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deployService",
            "variables": {
                "serviceId": serviceId
            },
            "query": f"""mutation 
                deployService($serviceId: String!) {{
                    deployService(serviceId: $serviceId) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "deployService")


def getServiceDeployment(self, deploymentId):
    fields = self.getTypeFields("ServiceDeployment")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getServiceDeployment",
            "variables": {
                "deploymentId": deploymentId
            },
            "query": f"""query 
                getServiceDeployment($deploymentId: String!) {{
                    getServiceDeployment(deploymentId: $deploymentId) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getServiceDeployment")

    
def addWorkspaceServices(self, workspaceId, serviceIds):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "addWorkspaceServices",
            "variables": {
                "workspaceId": workspaceId,
                "serviceIds": serviceIds
            },
            "query": """mutation 
                addWorkspaceServices($workspaceId: String!, $serviceIds: [String]!) {
                    addWorkspaceServices(workspaceId: $workspaceId, serviceIds: $serviceIds)
                }"""})
    return self.errorhandler(response, "addWorkspaceServices")


def removeWorkspaceServices(self, workspaceId, serviceIds):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "removeWorkspaceServices",
            "variables": {
                "workspaceId": workspaceId,
                "serviceIds": serviceIds
            },
            "query": """mutation 
                removeWorkspaceServices($workspaceId: String!, $serviceIds: [String]!) {
                    removeWorkspaceServices(workspaceId: $workspaceId, serviceIds: $serviceIds)
                }"""})
    return self.errorhandler(response, "removeWorkspaceServices")


def getServiceJobs(self, workspaceId, serviceId, limit=100, cursor=None, filters={}, fields=None):
    if fields is None: fields = self.getTypeFields("ServiceJob")
    fields = "\n".join(fields)
    if filters is None: filters = {}
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getServiceJobs",
            "variables": {
                "workspaceId": workspaceId,
                "serviceId": serviceId,
                "limit": limit,
                "cursor": cursor,
                "filters": filters
            },
            "query": F"""query 
                getServiceJobs($workspaceId: String!, $serviceId: String, $limit: Int, $cursor: String, $filters: ServiceJobsFilter) {{
                    getServiceJobs(workspaceId: $workspaceId, serviceId: $serviceId, limit: $limit, cursor: $cursor, filters: $filters) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getServiceJobs")


def createServiceJob(self, workspaceId, serviceId, name, description, payload):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createServiceJob",
            "variables": {
                "workspaceId": workspaceId,
                "serviceId": serviceId,
                "name": name,
                "description": description,
                "payload": payload
            },
            "query": """mutation 
                createServiceJob($workspaceId: String!, $serviceId: String!, $name: String!, $description: String, $payload: String!) {
                    createServiceJob(workspaceId: $workspaceId, serviceId: $serviceId, name: $name, description: $description, payload: $payload)
                }"""})
    return self.errorhandler(response, "createServiceJob")


def deleteServiceJob(self, jobId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteServiceJob",
            "variables": {
                "jobId": jobId
            },
            "query": """mutation 
                deleteServiceJob($jobId: String!) {
                    deleteServiceJob(jobId: $jobId)
                }"""})
    return self.errorhandler(response, "deleteServiceJob")


def getWorkspaceServiceCredentials(self, workspaceId, serviceId):
    fields = self.getTypeFields("ServiceCredentials")
    fields = "\n".join(fields)
    if serviceId is None:
        response = self.session.post(
            url = self.url, 
            headers = self.headers, 
            json = {
                "operationName": "getWorkspaceServiceCredentials",
                "variables": {
                    "workspaceId": workspaceId
                },
                "query": f"""query 
                    getWorkspaceServiceCredentials($workspaceId: String!) {{
                        getWorkspaceServiceCredentials(workspaceId: $workspaceId) {{
                            {fields}
                        }}
                    }}"""
            })
        return self.errorhandler(response, "getWorkspaceServiceCredentials")

    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getWorkspaceServiceCredentials",
            "variables": {
                "workspaceId": workspaceId,
                "serviceId": serviceId
            },
            "query": f"""query 
                getWorkspaceServiceCredentials($workspaceId: String!, $serviceId: String!) {{
                    getWorkspaceServiceCredentials(workspaceId: $workspaceId, serviceId: $serviceId) {{
                        {fields}
                    }}
                }}"""
        })
    return self.errorhandler(response, "getWorkspaceServiceCredentials")
    

def getInstanceTypes(self):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getInstanceTypes",
            "variables": {},
            "query": """query
                getInstanceTypes {
                    getInstanceTypes
                }"""
        })
    return self.errorhandler(response, "getInstanceTypes")


"""
Workspace Functions
"""

def set_workspace(self, workspaceId):
    """Set the workspace to the one you wish to work in.

    Parameters
    ----------
    workspaceId : str
        Workspace ID for the workspace you wish to work in.
    """
    from anatools.lib.print import print_color 
    self.check_logout()
    if workspaceId is None: raise Exception('WorkspaceId must be specified.')
    workspaces = self.get_workspaces(workspaceId=workspaceId)
    if len(workspaces) == 0: raise Exception(f'Workspace with workspaceId {workspaceId} not found!')
    self.workspace = workspaces[0]['workspaceId']
    self.organization = workspaces[0]['organizationId']
    if self.interactive: print_color(f'The current organization is: {self.organization}\nThe current workspace is: {self.workspace}', '91e600')
    return True


def get_workspaces(self, organizationId=None, workspaceId=None, cursor=None, limit=None, filters=None, fields=None):
    """Shows list of workspaces with id, name, and owner data.
    
    Parameters
    ----------
    organizationId : str
        Organization ID to filter on. Optional
    workspaceId : str
        Workspace ID to filter on. Optional
    cursor : str
        Cursor for pagination.
    limit : int
        Maximum number of workspaces to return.
    filters: dict
        Filters that limit output to entries that match the filter 
    fields : list
        List of fields to return, leave empty to get all fields.

    Returns
    -------
    list[dict]
        Workspace data for all workspaces for a user.
    """  
    if self.check_logout(): return
    if limit is not None and limit <= 100: items = limit
    else: items = 100
    workspaces = []
    while True:
        if limit and len(workspaces) + items > limit: items = limit - len(workspaces)
        ret = self.ana_api.getWorkspaces(organizationId=organizationId, workspaceId=workspaceId, limit=items, cursor=cursor, filters=filters, fields=fields)
        workspaces.extend(ret)
        if len(ret) < items or len(workspaces) == limit: break
        cursor = ret[-1]["workspaceId"]
    return workspaces


def create_workspace(self, name, description='', channelIds=[], volumeIds=[], code=None, tags=None, objective='', organizationId=None):
    """Create a new workspace with specific channels.
    
    Parameters
    ----------
    name : str    
        Workspace name.
    description : str
        Workspace description.
    channelIds : list[str]
        List of channel ids to add to workspace.
    volumeIds: list[str]
        List of volume ids that the workspace will have access to.
    code: str
        Content code that used for creating a workspace
    tags: list[str]
        List of tags to add to workspace.
    objective: str
        A goal to give the workspace.
    organizationId : str
        Organization ID. Defaults to current if not specified.  
    
    Returns
    -------
    str
        Workspace ID if creation was successful. Otherwise returns message.
    """    
    if self.check_logout(): return
    if name is None: raise ValueError("Name must be provided")
    if description is None: description = ''
    if code is None: code = ''
    if tags is None: tags = []
    if organizationId is None: organizationId = self.organization
    return self.ana_api.createWorkspace(organizationId=organizationId, name=name, description=description, channelIds=channelIds, volumeIds=volumeIds, code=code, tags=tags, objective=objective)


def delete_workspace(self, workspaceId=None):
    """Delete an existing workspace. 
    
    Parameters
    ----------
    workspaceId : str    
        Workspace ID for workspace to get deleted. Deletes current workspace if not specified. 
    
    Returns
    -------
    str
        Success or failure message if workspace was sucessfully removed.
    """
    if self.check_logout(): return
    if workspaceId is None: workspaceId = self.workspace 
    if self.interactive:
        response = input('This will destroy all data within the worksapce, inlcuding graphs, datasets and models.\nAre you certain you want to delete this workspace? (y/n)  ')
        if response.lower() != 'y': return True
    return self.ana_api.deleteWorkspace(workspaceId=workspaceId)


def edit_workspace(self, name=None, description=None, channelIds=None, volumeIds=None, ganIds=None, mapIds=None, tags=None, objective=None, workspaceId=None):
    """Edit workspace information. 
    
    Parameters
    ----------
    name : str    
        New name to replace old one.
    description : str
        New description to replace old one.
    channelIds: list[str]
        Names of channels that the workspace will have access to.
    volumeIds: list[str]
        List of volume ids that the workspace will have access to.
    ganIds: list[str]
        List of GAN ids that the workspace will have access to.
    mapIds: list[str]
        List of map ids that the workspace will have access to.
    tags: list[str]
        List of tags to add or remove.
    objective: str
        A goal to give the workspace.
    workspaceId : str    
        Workspace ID for workspace to update.
    
    Returns
    -------
    bool
        Success or failure message if workspace was sucessfully updated.
    """  
    if self.check_logout(): return
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.editWorkspace(workspaceId=workspaceId, name=name, description=description, channelIds=channelIds, volumeIds=volumeIds, ganIds=ganIds, mapIds=mapIds, tags=tags, objective=objective)


def mount_workspaces(self, workspaces):
    """Retrieves credentials for mounting workspaces.
    
    Parameters
    ----------
    workspaces : [str]
       Workspaces to retrieve mount credentials for.

    Returns
    -------
    dict
        Credential information.
    """
    self.check_logout()
    if not len(workspaces): raise Exception('The workspaces parameter must be a list of workspaceIds!')
    return self.ana_api.mountWorkspaces(workspaces=workspaces)


def create_workspace_with_template(self, templateId, organizationId=None):
    """Create a new workspace from a template.
    
    Parameters
    ----------
    templateId : str
        Template ID to create workspace from.
    """
    self.check_logout()
    if organizationId is None: organizationId = self.organization
    return self.ana_api.createWorkspaceWithTemplate(templateId=templateId, organizationId=organizationId)


def get_templates(self, organizationId=None):
    """Get all workspace templates.
    
    Parameters
    ----------
    organizationId : str
        Organization ID to get templates for.
    """
    self.check_logout()
    if organizationId is None: organizationId = self.organization
    return self.ana_api.getTemplates(organizationId=organizationId)


def create_template_request(self, filepath):
    """Create a new workspace template.
    
    Parameters
    ----------
    workspaceId : str
        Workspace ID to create workspace template from.
    filepath : str
        Path to the marketplace file.
    """
    import yaml
    import os

    self.check_logout()
    with open(filepath, 'r') as file: config = yaml.safe_load(file)
    workspaceId = os.environ.get('RENDEREDAI_WORKSPACE_ID', None)
    if workspaceId is None and config.get('workspaceId', None) is None: raise Exception('No workspace found.')
    return self.ana_api.createTemplateRequest(workspaceId=workspaceId, **config)


def get_template_requests(self, organizationId=None, templateId=None):
    """Get a workspace template request.
    
    Parameters
    ----------
    organizationId : str
        Organization ID to get template requests for.
    templateId : str
        Template ID to get template requests for.
    """
    self.check_logout()
    if organizationId is None: organizationId = self.organization
    return self.ana_api.getTemplateRequests(organizationId=organizationId, templateId=templateId)



"""
API Keys calls.
"""

def getAPIKeyContext(self, apiKey):
    fields = self.getTypeFields("APIKey")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getAPIKeyContext",
            "variables": {
                "apiKey": apiKey
            },
            "query": f"""query getAPIKeyContext($apiKey: String!) {{
                            getAPIKeyContext(apiKey: $apiKey) {{
                                {fields}
                            }}
                        }}"""})
    return self.errorhandler(response, "getAPIKeyContext")

def getCurrentUserContext(self):
    """Fetches the current user's context using the Bearer token in session headers."""
    import time
    response = self.session.post(
        url=self.url,
        headers=self.headers, # Assumes Bearer token is already in self.headers
        json={
            "operationName": "GetCurrentUser",
            "variables": {},
            "query": """query GetCurrentUser {
                            getCurrentUser { # Or the actual query name and returned fields
                                userId
                                name # Added name as per definition
                                email
                                # Add other fields like defaultOrganizationId, defaultWorkspaceId if available
                            }
                        }"""})
    # The errorhandler should manage API errors, including token-related ones.
    data = self.errorhandler(response, "getCurrentUser")
    if data and 'userId' in data: # Basic validation
        # If the token itself has an expiry and it's returned by an introspection endpoint,
        # that would be ideal. For now, we don't have 'expiresAt' from this call.
        # The idtoken will be the bearer token itself, which anaclient.py will set.
        pass # Data is good
    return data


def getAPIKeys(self, apiKey=None, name=None):
    fields = self.getTypeFields("APIKey")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getAPIKeys",
            "variables": {
                "apiKey": apiKey,
                "name": name
            },
            "query": f"""query getAPIKeys ($apiKey: String, $name: String) {{
                            getAPIKeys(apiKey: $apiKey, name: $name) {{
                                {fields}
                            }}
                        }}"""})
    return self.errorhandler(response, "getAPIKeys")


def createAPIKey(self, name, scope="user", organizationId=None, workspaceId=None, expires=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createAPIKey",
            "variables": {
                "name": name,
                "scope": scope,
                "organizationId": organizationId,
                "workspaceId": workspaceId,
                "expiresAt": expires
            },
            "query": """mutation createAPIKey($name: String!, $scope: String, $organizationId: String, $workspaceId: String, $expiresAt: String) {
                            createAPIKey(name: $name, scope: $scope, organizationId: $organizationId, workspaceId: $workspaceId, expiresAt: $expiresAt)
                        }"""})
    return self.errorhandler(response, "createAPIKey")


def deleteAPIKey(self, name):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteAPIKey",
            "variables": {
                "name": name
            },
            "query": """mutation deleteAPIKey($name: String!) {
                            deleteAPIKey(name: $name)
                        }"""})
    return self.errorhandler(response, "deleteAPIKey")



"""
Workspaces API calls.
"""

def getWorkspaces(self, organizationId=None, workspaceId=None, cursor=None, limit=100, filters={}, fields=None):
    if fields is None: fields = self.getTypeFields("Workspace")
    fields = "\n".join(fields)
    if filters is None: filters = {}
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getWorkspaces",
            "variables": {
                "organizationId": organizationId,
                "workspaceId": workspaceId,
                "limit": limit,
                "cursor": cursor,
                "filters": filters
            },
            "query": f"""query 
                getWorkspaces($organizationId: String, $workspaceId: String, $limit: Int, $cursor: String, $filters: WorkspaceFilter) {{
                    getWorkspaces(organizationId: $organizationId, workspaceId: $workspaceId, limit: $limit, cursor: $cursor, filters: $filters) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getWorkspaces")


def createWorkspace(self, organizationId, name, description='', channelIds=[], volumeIds=[], code=None, tags=[], objective=''):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createWorkspace",
            "variables": {
                "organizationId": organizationId,
                "name": name,
                "description": description,
                "channelIds": channelIds,
                "volumeIds": volumeIds,
                "code": code,
                "tags": tags,
                "objective": objective
            },
            "query": """mutation 
                createWorkspace($organizationId: String!, $name: String!, $description: String, $channelIds: [String]!, $volumeIds: [String]!, $code: String!, $tags: [String], $objective: String) {
                    createWorkspace(organizationId: $organizationId, name: $name, description: $description, channelIds: $channelIds, volumeIds: $volumeIds, code: $code, tags: $tags, objective: $objective)
                }"""})
    return self.errorhandler(response, "createWorkspace")


def deleteWorkspace(self, workspaceId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteWorkspace",
            "variables": {
                "workspaceId": workspaceId
            },
            "query": """mutation 
                deleteWorkspace($workspaceId: String!) {
                    deleteWorkspace(workspaceId: $workspaceId)
                }"""})
    return self.errorhandler(response, "deleteWorkspace")


def editWorkspace(self, workspaceId, name=None, description=None, channelIds=None, volumeIds=None, ganIds=None, mapIds=None, tags=[], objective=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editWorkspace",
            "variables": {
                "workspaceId": workspaceId,
                "name": name,
                "description": description,
                "channelIds": channelIds,
                "volumeIds": volumeIds,
                "ganIds": ganIds,
                "mapIds": mapIds,
                "tags": tags,
                "objective": objective
            },
            "query": """mutation 
                editWorkspace($workspaceId: String!, $name: String, $description: String, $channelIds: [String], $volumeIds: [String], $ganIds: [String], $mapIds: [String], $tags: [String], $objective: String) {
                    editWorkspace(workspaceId: $workspaceId, name: $name, description: $description, channelIds: $channelIds, volumeIds: $volumeIds, ganIds: $ganIds, mapIds: $mapIds, tags: $tags, objective: $objective)
                }"""})
    return self.errorhandler(response, "editWorkspace")


def mountWorkspaces(self, workspaces):
    fields = self.getTypeFields("WorkspaceCredentials")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "mountWorkspaces",
            "variables": {
                "workspaces": workspaces
            },
            "query": f"""mutation 
                mountWorkspaces($workspaces: [String]!) {{
                    mountWorkspaces(workspaces: $workspaces) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "mountWorkspaces")



def createWorkspaceWithTemplate(self, templateId, organizationId=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createWorkspaceWithTemplate",
            "variables": {
                "templateId": templateId,
                "organizationId": organizationId
            },
            "query": """mutation 
                createWorkspaceWithTemplate($templateId: String!, $organizationId: String) {
                    createWorkspaceWithTemplate(templateId: $templateId, organizationId: $organizationId)
                }"""})
    return self.errorhandler(response, "createWorkspaceWithTemplate")


def getTemplates(self, organizationId=None, fields=None):
    if fields is None: fields = self.getTypeFields("WorkspaceTemplate")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getTemplates",
            "variables": {
                "organizationId": organizationId
            },
            "query": f"""query 
                getTemplates($organizationId: String) {{
                    getTemplates(organizationId: $organizationId) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getTemplates")


def createTemplateRequest(self, workspaceId, name=None, description=None, documentation=None, organizationIds=None, instance=None, index=None, demoVideo=None, featured=None, keyFeatures=None, license=None, tags=None, thumbnail=None, version=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createTemplateRequest",
            "variables": {
                "workspaceId": workspaceId,
                "name": name,
                "description": description,
                "documentation": documentation,
                "organizationIds": organizationIds,
                "instance": instance,
                "index": index,
                "demoVideo": demoVideo,
                "featured": featured,
                "keyFeatures": keyFeatures,
                "license": license,
                "tags": tags,
                "thumbnail": thumbnail,
                "version": version
            },
            "query": """mutation 
                createTemplateRequest($workspaceId: String!, $name: String, $description: String, $documentation: String, $organizationIds: [String], $instance: String, $index: Int, $demoVideo: String, $featured: Boolean, $keyFeatures: [String], $license: String, $tags: [String], $thumbnail: String, $version: Int) {
                    createTemplateRequest(workspaceId: $workspaceId, name: $name, description: $description, documentation: $documentation, organizationIds: $organizationIds, instance: $instance, index: $index, demoVideo: $demoVideo, featured: $featured, keyFeatures: $keyFeatures, license: $license, tags: $thumbnail, version: $version)
                }"""})
    return self.errorhandler(response, "createTemplateRequest")


def getTemplateRequests(self, organizationId, templateId=None, fields=None):
    if fields is None: fields = self.getTypeFields("TemplateRequests")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getTemplateRequests",
            "variables": {
                "organizationId": organizationId,
                "templateId": templateId
            },
            "query": f"""query 
                getTemplateRequests($organizationId: String!, $templateId: String) {{
                    getTemplateRequests(organizationId: $organizationId, templateId: $templateId) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getTemplateRequests")



"""
GAN API calls.
"""

def getGANModels(self, organizationId, workspaceId, modelId, limit=100, cursor=None, filters={}, fields=None):
    if fields is None: fields = self.getTypeFields("GANModel")
    fields = "\n".join(fields)
    if filters is None: filters = {}
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getGANModels",
            "variables": {
                "organizationId": organizationId,
                "workspaceId": workspaceId,
                "modelId": modelId,
                "limit": limit,
                "cursor": cursor,
                "filters": filters
            },
            "query": f"""query 
                getGANModels($organizationId: String, $workspaceId: String, $modelId: String, $limit: Int, $cursor: String, $filters: GANModelFilter) {{
                    getGANModels(organizationId: $organizationId, workspaceId: $workspaceId, modelId: $modelId, limit: $limit, cursor: $cursor, filters: $filters){{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getGANModels")


def getGANDatasets(self, datasetId, workspaceId, gandatasetId, limit=100, cursor=None, fields=None):
    if fields is None: fields = self.getTypeFields("GANDataset")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getGANDatasets",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "gandatasetId": gandatasetId,
                "limit": limit,
                "cursor": cursor
            },
            "query": f"""query 
                getGANDatasets($workspaceId: String!, $datasetId: String, $gandatasetId: String, $limit: Int, $cursor: String) {{
                    getGANDatasets(workspaceId: $workspaceId, datasetId: $datasetId, gandatasetId: $gandatasetId, limit: $limit, cursor: $cursor) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getGANDatasets")


def createGANDataset(self, workspaceId, datasetId, modelId, name, description='', tags=[]):
    if tags is None: tags = []
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createGANDataset",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId,
                "modelId": modelId,
                "name": name,
                "description": description,
                "tags": tags
            },
            "query": """mutation 
                createGANDataset($workspaceId: String!, $datasetId: String!, $modelId: String!, $name: String, $description: String, $tags: [String]) {
                    createGANDataset(workspaceId: $workspaceId, datasetId: $datasetId, modelId: $modelId, name: $name, description: $description, tags: $tags)
                }"""})
    return self.errorhandler(response, "createGANDataset")


def deleteGANDataset(self, datasetId, workspaceId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteGANDataset",
            "variables": {
                "workspaceId": workspaceId,
                "datasetId": datasetId
            },
            "query": """mutation 
                deleteGANDataset($workspaceId: String!, $datasetId: String!) {
                    deleteGANDataset(workspaceId: $workspaceId, datasetId: $datasetId)
                }"""})
    return self.errorhandler(response, "deleteGANDataset")


def uploadGANModel(self, organizationId, name, size, description="", flags='', tags=[]):
    fields = self.getTypeFields("UploadGANResponse")
    fields = "\n".join(fields)
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "uploadGANModel",
            "variables": {
                "organizationId": organizationId,
                "name": name,
                "size": size,
                "description": description,
                "flags": flags,
                "tags": tags
            },
            "query": f"""mutation 
                uploadGANModel($organizationId: String!, $name: String!, $size: Float!, $description: String, $flags: String, $tags: [String]) {{
                    uploadGANModel(organizationId: $organizationId, name: $name, size: $size, description: $description, flags: $flags, tags: $tags) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "uploadGANModel")


def uploadGANModelFinalizer(self, uploadId, key, parts):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "uploadGANModelFinalizer",
            "variables": {
                "uploadId": uploadId,
                "key": key,
                "parts": parts,
            },
            "query": """mutation 
                uploadGANModelFinalizer($uploadId: String!, $key: String!, $parts: [MultipartInput!]) {
                    uploadGANModelFinalizer(uploadId: $uploadId, key: $key, parts: $parts)
                }"""})
    return self.errorhandler(response, "uploadGANModelFinalizer")


def deleteGANModel(self, modelId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteGANModel",
            "variables": {
                "modelId": modelId,
            },
            "query": """mutation 
                deleteGAN($modelId: String!) {
                    deleteGAN(modelId: $modelId)
                }"""})
    return self.errorhandler(response, "deleteGANModel")


def editGANModel(self, modelId, name=None, description=None, flags=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editGAN",
            "variables": {
                "modelId": modelId,
                "name": name,
                "description": description,
                "flags": flags
            },
            "query": """mutation 
                editGAN($modelId: String!, $name: String, $description: String, $flags: String) {
                    editGAN(modelId: $modelId, name: $name, description: $description, flags: $flags)
                }"""})
    return self.errorhandler(response, "editGAN")


def downloadGANModel(self, modelId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "downloadGAN",
            "variables": {
                "modelId": modelId
            },
            "query": """mutation 
                downloadGAN($modelId: String!) {
                    downloadGAN(modelId: $modelId) 
                }"""})
    return self.errorhandler(response, "downloadGAN")



import pytest


### Fixtures ###
@pytest.fixture(scope="session")
def add_organization_member_fixture(client):
    email = "test@rendered.ai"
    client.add_organization_member(organizationId=client.organization, email=email, role="member")
    return email


### Tests ###
def test_set_organization(client):
    organizationId = client.organization
    workspaceId = client.workspace
    status = client.set_organization(organizationId=organizationId)
    assert status == True
    assert client.organization == organizationId
    status = client.set_workspace(workspaceId=workspaceId)
    assert status == True
    assert client.workspace == workspaceId

def test_get_organizations(client):
    organizations = client.get_organizations()
    assert isinstance(organizations, list)
    assert len(organizations) > 0

def test_edit_organization(client):
    organization = client.get_organizations(organizationId=client.organization)[0]
    status = client.edit_organization(organizationId=organization['organizationId'], name="anatools-edit-organization")
    assert status == True
    organizations = client.get_organizations(organizationId=client.organization)
    assert isinstance(organizations, list)
    assert len(organizations) == 1
    assert organizations[0]['name'] == "anatools-edit-organization"
    status = client.edit_organization(organizationId=organization['organizationId'], name=organization['name'])
    assert status == True
    
def test_get_organization_members(client):
    members = client.get_organization_members(organizationId=client.organization)
    assert isinstance(members, list)
    assert len(members) > 0

def test_get_organization_invites(client, add_organization_member_fixture):
    invites = client.get_organization_invites(organizationId=client.organization)
    assert isinstance(invites, list)
    userinvites = [invite for invite in invites if invite['email'] == add_organization_member_fixture]
    assert len(userinvites) == 1

@pytest.mark.dependency(depends=["test_get_organization_invites"])
def test_remove_organization_invitation(client, add_organization_member_fixture):
    email = add_organization_member_fixture
    status = client.remove_organization_invitation(organizationId=client.organization, email=email)
    assert status == True

# def test_remove_organization_member(client, add_organization_member_fixture):
#     email = add_organization_member_fixture
#     status = client.remove_organization_member(organizationId=client.organization, email=email)
#     assert status == True

# def test_edit_organization_member(client, add_organization_member_fixture):
#     email = add_organization_member_fixture
#     status = client.edit_organization_member(organizationId=client.organization, email=email, role="member")
#     assert status == True
#     members = client.get_organization_members(organizationId=client.organization)
#     usermembers = [member for member in members if member['email'] == email]
#     assert len(usermembers) == 1
#     assert usermembers[0]['role'] == 'member'    



import json
import os
import datetime
from PIL import Image


def create_cocodata():
    cocodata = dict()
    cocodata["info"] = {
        "description": "Rendered.AI Synthetic Dataset",
        "url": "https://rendered.ai/",
        "contributor": "info@rendered.ai",
        "version": "1.0",
        "year": str(datetime.datetime.now().year),
        "date_created": datetime.datetime.now().isoformat(),
    }
    cocodata["licenses"] = [{"id": 0, "url": "https://rendered.ai/", "name": "Rendered.AI License"}]
    cocodata["images"] = list()
    cocodata["categories"] = list()
    cocodata["annotations"] = list()
    return cocodata


def index_metadata_by_filename(metadir):
    """Pre-index all metadata files by their filename field for O(1) lookup"""
    metadata_index = {}

    for mf in os.listdir(metadir):
        if not mf.endswith(".json"):
            continue
        try:
            filepath = os.path.join(metadir, mf)
            with open(filepath, "r") as f:
                metadata = json.load(f)
                filename = metadata.get("filename")
                if filename:
                    metadata_index[filename] = metadata
        except (json.JSONDecodeError, IOError):
            continue

    return metadata_index


def get_image_dimensions(imgpath):
    """Cache image dimensions to avoid repeated file opens"""
    try:
        with Image.open(imgpath) as im:
            return im.size
    except:
        return None, None


def process_annotation_file(ann_file_path, metadata_index, imgdir, mapping=None):
    """Process a single annotation file and return the annotations and image data"""
    with open(ann_file_path, "r") as f:
        anns = json.load(f)

    image_filename = anns.get("filename")
    if not image_filename:
        return None, None, None

    metadata = metadata_index.get(image_filename)
    if metadata is None:
        return None, None, None

    annotations = []
    categories = {}

    for obj in metadata["objects"]:
        if mapping is None:
            for ann in anns["annotations"]:
                if ann["id"] == obj["id"]:
                    obj_type = obj["type"]
                    annotations.append(
                        {
                            "obj_id": ann["id"],
                            "obj_type": obj_type,
                            "segmentation": ann["segmentation"],
                            "bbox": ann["bbox"],
                            "area": ann["bbox"][2] * ann["bbox"][3],
                        }
                    )
                    break
        else:
            for prop in mapping["properties"]:
                if eval(prop):
                    for ann in anns["annotations"]:
                        if ann["id"] == obj["id"]:
                            class_num = mapping["properties"][prop]
                            categories[class_num] = mapping["classes"][class_num]
                            annotations.append(
                                {
                                    "obj_id": ann["id"],
                                    "category_id": class_num,
                                    "segmentation": ann["segmentation"],
                                    "bbox": ann["bbox"],
                                    "area": ann["bbox"][2] * ann["bbox"][3],
                                }
                            )
                            break

    # Prepare image data
    date = metadata.get("date", datetime.datetime.now().isoformat())
    imgdata = {"file_name": metadata["filename"], "date_captured": date, "license": 0}

    # Get image dimensions
    try:
        imgdata["width"] = metadata["sensor"]["resolution"][0]
        imgdata["height"] = metadata["sensor"]["resolution"][1]
    except:
        width, height = get_image_dimensions(os.path.join(imgdir, image_filename))
        if width and height:
            imgdata["width"] = width
            imgdata["height"] = height

    return annotations, imgdata, categories


def convert_coco(datadir, outdir, mapping=None):
    """Optimized COCO conversion with pre-indexing and optional parallel processing"""

    annsdir = os.path.join(datadir, "annotations")
    metadir = os.path.join(datadir, "metadata")
    imgdir = os.path.join(datadir, "images")

    # Pre-index metadata for O(1) lookup
    print("Indexing metadata files...")
    metadata_index = index_metadata_by_filename(metadir)
    print(f"Indexed {len(metadata_index)} metadata files")

    # Get sorted list of annotation files
    annsfiles = sorted([f for f in os.listdir(annsdir) if f.endswith(".json")])

    cocodata = create_cocodata()
    cats = {0: [None, "coco_background"]}
    all_annotations = []
    all_images = []

    imgid = 0
    annid = 0

    # Sequential processing
    for f in annsfiles:
        annotations, imgdata, categories = process_annotation_file(
            os.path.join(annsdir, f), metadata_index, imgdir, mapping
        )

        if annotations and imgdata:
            if mapping is None:
                for ann in annotations:
                    obj_type = ann["obj_type"]
                    if [None, obj_type] in cats.values():
                        class_num = [k for k, v in cats.items() if v == [None, obj_type]][0]
                    else:
                        class_num = len(cats.keys())
                        cats[class_num] = [None, obj_type]

                    annotation = {
                        "id": annid,
                        "image_id": imgid,
                        "category_id": class_num,
                        "segmentation": ann["segmentation"],
                        "area": ann["area"],
                        "bbox": ann["bbox"],
                        "iscrowd": 0,
                    }
                    all_annotations.append(annotation)
                    annid += 1
            else:
                cats.update(categories)
                for ann in annotations:
                    annotation = {
                        "id": annid,
                        "image_id": imgid,
                        "category_id": ann["category_id"],
                        "segmentation": ann["segmentation"],
                        "area": ann["area"],
                        "bbox": ann["bbox"],
                        "iscrowd": 0,
                    }
                    all_annotations.append(annotation)
                    annid += 1

            imgdata["id"] = imgid
            all_images.append(imgdata)
            imgid += 1

    # Build categories list
    sorted_cats = dict(sorted(cats.items()))
    for class_num, cat in sorted_cats.items():
        if class_num == 0:
            continue
        cocodata["categories"].append({"id": class_num, "name": cat[-1], "supercategory": cat[0]})

    # Add all data to cocodata
    cocodata["annotations"] = all_annotations
    cocodata["images"] = all_images

    # Write to file at once
    output_path = os.path.join(outdir, "coco.json")
    with open(output_path, "w") as f:
        json.dump(cocodata, f)

    print(f"Conversion complete: {len(all_images)} images, {len(all_annotations)} annotations")
    return cocodata



"""
API Keys Functions
"""


def get_api_keys(self):
    """Queries the api keys associated with user's account. This call will return data only when logged in with email/password.
    
    Parameters
    ----------
    
    Returns
    -------
    [dict]
        Names of API keys associated with user's account.
    """
    if self.check_logout(): return
    return self.ana_api.getAPIKeys()


def create_api_key(self, name, scope='user', organizationId=None, workspaceId=None, expires=None):
    """Creates a new API Key for the User; the key will only be visible once, so make sure to save it.
        To use the API Key on login, ensure it is set as an environment variable called RENDEREDAI_API_KEY or with the APIKey parameter when initializing the anatools client.
        This call can only be used when logged in with email/password.
    
    Parameters
    ----------
    name: str
        Name of the API Key. 
    scope: str
        Scope of the API Key, this can be set to 'user', 'organization', or 'workspace' to limit the scope of access for the API Key. 
    organizationId: str
        Organization ID to set the API Key access to a particular organization's data.
    workspaceId: str
        Workspace ID to set the API Key access to a particular workspace's data.
    expires : str
        Expiration date to set for the API Key. If no expiration is provided, the key will not expire. 
        
    Returns
    -------
    str
        API Key
    """
    if self.check_logout(): return
    if self.user is None: raise Exception("Please sign in with username/password to create an API Key.")
    if name is None: raise Exception("Name must be set for an API Key, each API Key name must be unique.")
    if scope not in ['user', 'organization', 'workspace']: raise Exception("Scope must be set to 'user', 'organization', or 'workspace'.")
    if organizationId and organizationId not in [org['organizationId'] for org in self.get_organizations()]: raise Exception("Organization ID not found.")
    if workspaceId and workspaceId not in [workspace['workspaceId'] for workspace in self.get_workspaces()]: raise Exception("Workspace ID not found.")
    return self.ana_api.createAPIKey(name=name, scope=scope, organizationId=organizationId, workspaceId=workspaceId, expires=expires)


def delete_api_key(self, name):
    """Deletes the API key from user account. This call can only be used when logged in with email/password.
    
    Parameters
    ----------
    name: str
        Name of the API Key to delete.
        
    Returns
    -------
    bool
        Success or failure message about API key deletion
    """
    if self.check_logout(): return
    if name is None: return
    return self.ana_api.deleteAPIKey(name=name)



import time
import os
import pytest
from .graphs_test import upload_graph_fixture, get_default_graph_fixture

### Fixtures ###
@pytest.fixture(name="create_dataset_fixture", scope="session")
def create_dataset_fixture(client, upload_graph_fixture):
    graphId = upload_graph_fixture
    datasetId = client.create_dataset(name=f"anatools-dataset-create", graphId=graphId, description="anatools test dataset", runs=100, priority=1, seed=1, tags=["test"], workspaceId=client.workspace)
    while client.get_datasets(datasetId=datasetId)[0]['status'] not in ['success', 'failed', 'cancelled', 'timeout']: time.sleep(10)
    return datasetId

@pytest.fixture(name="download_dataset_fixture", scope="session")
def download_dataset_fixture(client, create_dataset_fixture):
    datasetId = create_dataset_fixture
    datasetfile = client.download_dataset(datasetId=datasetId, workspaceId=client.workspace)
    return datasetfile

@pytest.fixture(name="get_dataset_files_fixture", scope="session")
def get_dataset_files_fixture(client, create_dataset_fixture):
    datasetId = create_dataset_fixture
    files = client.get_dataset_files(datasetId=datasetId, workspaceId=client.workspace)
    return files

@pytest.fixture(name="upload_dataset_fixture", scope="session")
def upload_dataset_fixture(client, download_dataset_fixture):
    datasetfile = download_dataset_fixture
    datasetId = client.upload_dataset(name=f"anatools-dataset-upload", file=datasetfile, workspaceId=client.workspace)
    return datasetId
                
@pytest.fixture(name="create_mixed_dataset_fixture", scope="session")
def create_mixed_dataset_fixture(client, create_dataset_fixture, upload_dataset_fixture):
    datasetId1 = create_dataset_fixture
    datasetId2 = upload_dataset_fixture
    mixedDatasetId = client.create_mixed_dataset(name=f"anatools-dataset-mixed", parameters={datasetId1: {"samples": 100}, datasetId2: {"samples": 100}}, description="anatools test mixed dataset", seed=1, tags=["test"], workspaceId=client.workspace)
    return mixedDatasetId

@pytest.fixture(name="get_dataset_run_fixture", scope="session")
def get_dataset_run_fixture(client, create_dataset_fixture):
    datasetId = create_dataset_fixture
    runs = client.get_dataset_runs(datasetId=datasetId, workspaceId=client.workspace)
    return runs[-1]['runId']


### Tests ###
@pytest.mark.smoke
def test_get_datasets(client):
    datasets = client.get_datasets(workspaceId=client.workspace)
    assert isinstance(datasets, list)

@pytest.mark.smoke
def test_create_dataset(client, create_dataset_fixture):
    datasetId = create_dataset_fixture
    assert isinstance(datasetId, str)
    datasets = client.get_datasets(datasetId=datasetId)
    assert len(datasets) == 1
    assert datasets[0]['status'] == 'success'

def test_download_dataset(download_dataset_fixture):
    datasetfile = download_dataset_fixture
    assert os.path.isfile(datasetfile)

def test_edit_dataset(client, create_dataset_fixture):
    datasetId = create_dataset_fixture
    status = client.edit_dataset(datasetId=datasetId, name="anatools-dataset-edit", workspaceId=client.workspace)
    assert status == True

@pytest.mark.smoke
@pytest.mark.dependency(depends=["test_get_datasets"])
@pytest.mark.dependency(depends=["test_edit_dataset"], optional=True)
def test_delete_dataset(client, create_dataset_fixture):
    datasetId = create_dataset_fixture
    status = client.delete_dataset(datasetId=datasetId, workspaceId=client.workspace)
    assert status == True

def test_cancel_dataset(client, upload_graph_fixture):
    graphId = upload_graph_fixture
    datasetId = client.create_dataset(name="anatools-dataset-cancel", graphId=graphId, description="anatools test cancel dataset", runs=1, priority=1, seed=1, tags=["test"], workspaceId=client.workspace)
    assert isinstance(datasetId, str)
    while client.get_datasets(datasetId=datasetId)[0]['status'] not in ['running', 'starting']: time.sleep(10)
    status = client.cancel_dataset(datasetId=datasetId, workspaceId=client.workspace)
    assert status == True
    while client.get_datasets(datasetId=datasetId)[0]['status'] not in ['cancelled']: time.sleep(10)
    assert client.get_datasets(datasetId=datasetId)[0]['status'] == 'cancelled'

def test_upload_dataset(upload_dataset_fixture):
    datasetId = upload_dataset_fixture
    assert isinstance(datasetId, str)

def test_create_mixed_dataset(create_mixed_dataset_fixture):
    datasetId = create_mixed_dataset_fixture
    assert isinstance(datasetId, str)

def test_get_dataset_files(get_dataset_files_fixture):
    files = get_dataset_files_fixture
    assert isinstance(files, list)
    assert len(files) > 0

def test_get_dataset_runs(get_dataset_run_fixture):
    runId = get_dataset_run_fixture
    assert isinstance(runId, str)

def test_get_dataset_log(client, create_dataset_fixture, get_dataset_run_fixture):
    datasetId = create_dataset_fixture
    runId = get_dataset_run_fixture
    log = client.get_dataset_log(datasetId=datasetId, runId=runId, workspaceId=client.workspace)
    assert isinstance(log, dict)



import os
import pytest
import time

# TODO: support remote development and SSH key tests

# ### Fixtures ###
# @pytest.fixture(scope="session")
# def create_remote_development_fixture(client):
#     editorId = client.create_remote_development(channelId=client.channel)
#     while client.get_remote_development(editorId=editorId)['status'] not in ['running', 'failed', 'cancelled', 'timeout']: time.sleep(10)
#     return editorId

# @pytest.fixture(scope="session")
# def register_ssh_key_fixture(filename=None):
#     filename = "test-ssh-key.pub"
#     with open(filename, "w") as f: f.write("test")
#     client.register_ssh_key(filename=filename)
#     os.remove(filename)
#     return filename


# ### Tests ###
# def test_list_remote_development(client):
#     editors = client.list_remote_development()
#     assert isinstance(editors, list)

# def test_create_remote_development(client, create_remote_development_fixture):
#     editorId = create_remote_development_fixture
#     assert client.get_remote_development(editorId=editorId)['status'] == ['running']

# @pytest.mark.dependency(depends=["stop_remote_development"])
# def test_delete_remote_development(client, create_remote_development_fixture):
#     editorId = create_remote_development_fixture
#     status = client.delete_remote_development(editorId=editorId)
#     assert status == True

# def test_stop_remote_development(client, create_remote_development_fixture):
#     editorId = create_remote_development_fixture
#     status = client.stop_remote_development(editorId=editorId)
#     assert status == True

# def test_start_remote_development(client, create_remote_development_fixture):
#     editorId = create_remote_development_fixture
#     status = client.start_remote_development(editorId=editorId)
#     assert status == True

# def test_prepare_ssh_remote_development(client, create_remote_development_fixture):
#     editorId = create_remote_development_fixture
#     status = client.prepare_ssh_remote_development(editorId=editorId)
#     assert status == True   
   
# def test_get_ssh_keys(client, register_ssh_key_fixture):
#     sshkeys = client.get_ssh_keys()
#     assert isinstance(sshkeys, list)
#     assert len(sshkeys) > 0

# @pytest.mark.dependency(depends=["test_get_ssh_keys"])
# def test_deregister_ssh_key(client, register_ssh_key_fixture):
#     sshkey = register_ssh_key_fixture
#     status = client.deregister_ssh_key(sshkey=sshkey)
#     assert status == True
   


"""
GAN API calls.
"""

def getUMAPs(self, umapId, datasetId, workspaceId, limit=100, cursor=None, filters={}, fields=None):
    if fields is None: fields = self.getTypeFields("UMAP")
    fields = "\n".join(fields)
    if filters is None: filters = {}
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "getUMAPs",
            "variables": {
                "workspaceId": workspaceId,
                "umapId": umapId,
                "datasetId": datasetId,
                "limit": limit,
                "cursor": cursor,
                "filters": filters
            },
            "query": f"""query 
                getUMAPs($workspaceId: String!, $datasetId: String, $umapId: String, $limit: Int, $cursor: String, $filters: UMAPFilter) {{
                    getUMAPs(workspaceId: $workspaceId, datasetId: $datasetId, umapId: $umapId, limit: $limit, cursor: $cursor, filters: $filters) {{
                        {fields}
                    }}
                }}"""})
    return self.errorhandler(response, "getUMAPs")


def createUMAP(self, workspaceId, name, datasetIds, samples, description=None, seed=None, tags=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "createUMAP",
            "variables": {
                "workspaceId": workspaceId,
                "name": name,
                "datasetIds": datasetIds,
                "samples": samples,
                "description": description,
                "seed": seed,
                "tags": tags
            },
            "query": """mutation 
                createUMAP($workspaceId: String!, $name: String, $datasetIds: [String]!, $samples: [Int]!, $description: String, $seed: Int, $tags: [String]) {
                    createUMAP(workspaceId: $workspaceId, name: $name, datasetIds: $datasetIds, samples: $samples, description: $description, seed: $seed, tags: $tags)
                }"""})
    return self.errorhandler(response, "createUMAP")


def deleteUMAP(self, umapId, workspaceId):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "deleteUMAP",
            "variables": {
                "workspaceId": workspaceId,
                "umapId": umapId,
            },
            "query": """mutation 
                deleteUMAP($workspaceId: String!, $umapId: String!) {
                    deleteUMAP(workspaceId: $workspaceId, umapId: $umapId)
                }"""})
    return self.errorhandler(response, "deleteUMAP")


def editUMAP(self, workspaceId, umapId, name=None, description=None, tags=None):
    response = self.session.post(
        url = self.url, 
        headers = self.headers, 
        json = {
            "operationName": "editUMAP",
            "variables": {
                "workspaceId": workspaceId,
                "umapId": umapId,
                "name": name,
                "description": description,
                "tags": tags
            },
            "query": """mutation 
                editUMAP($workspaceId: String!, $umapId: String!, $name: String, $description: String, $tags: [String]) {
                    editUMAP(workspaceId: $workspaceId, umapId: $umapId, name: $name, description: $description, tags: $tags)
                }"""})
    return self.errorhandler(response, "editUMAP")


import json
import os

def convert_classification(datadir, outdir, mapping=None):

    annsdir = os.path.join(datadir, "annotations")
    metadir = os.path.join(datadir, "metadata")
    imgdir = os.path.join(datadir, "images")
    classification = {}

    for img in os.listdir(imgdir):
        imgbase, imgext = os.path.splitext(img)
        metadata = json.load(open(os.path.join(metadir, imgbase + "-metadata.json")))
        anns = json.load(open(os.path.join(annsdir, imgbase + "-ana.json")))
        for obj in metadata['objects']:
            if mapping is None:
                for ann in anns['annotations']:
                    if ann['id'] == obj['id']:
                        class_name = obj['type']
                        if class_name not in classification: classification[class_name] = []
                        if img not in classification[class_name]: classification[class_name].append(img)
            else:
                for prop in mapping['properties']:
                    if eval(prop):
                        for ann in anns['annotations']:
                            if ann['id'] == obj['id']: 
                                class_num = mapping['properties'][prop]
                                class_name = mapping['classes'][class_num][-1]
                                if class_name not in classification: classification[class_name] = []
                                if img not in classification[class_name]: classification[class_name].append(img)

    with open(os.path.join(outdir,'classification.json'), 'w+') as cf:
        json.dump(classification, cf)




import os
import pytest
import time
from .datasets_test import create_dataset_fixture

# # TODO: support ml tests

# ### Fixtures ###
# @pytest.fixture(scope="session")
# def get_ml_architectures_fixture(client):
#     architectures = client.get_ml_architectures()
#     return architectures

# @pytest.fixture(scope="session")
# def create_ml_model_fixture(client, get_ml_architectures_fixture, create_dataset_fixture):
#     architectureId = get_ml_architectures_fixture[0]['architectureId']
#     datasetId = create_dataset_fixture
#     parameters = {}
#     modelId = client.create_ml_model(datasetId=datasetId, architectureId=architectureId, name='anatools-ml-model', parameters=parameters, workspaceId=client.workspace)
#     while client.get_ml_models(modelId=modelId, workspaceId=client.workspace)[0]['status'] not in ['success', 'failed', 'cancelled', 'timeout']: time.sleep(10)
#     return modelId

# @pytest.fixture(scope="session")
# def download_ml_model_fixture(client, create_ml_model_fixture):
#     modelId = create_ml_model_fixture
#     modelFile = client.download_ml_model(modelId=modelId, workspaceId=client.workspace)
#     return modelFile

# @pytest.fixture(scope="session")
# def create_ml_inference_fixture(client, create_ml_model_fixture, create_dataset_fixture):
#     modelId = create_ml_model_fixture
#     datasetId = create_dataset_fixture
#     inferenceId = client.create_ml_inference(modelId=modelId, datasetId=datasetId, workspaceId=client.workspace)
#     while client.get_ml_inferences(inferenceId=inferenceId, workspaceId=client.workspace)[0]['status'] not in ['success', 'failed', 'cancelled', 'timeout']: time.sleep(10)
#     return inferenceId

# ### Tests ###
# def test_get_ml_architectures(get_ml_architectures_fixture):
#     architectures = get_ml_architectures_fixture
#     assert isinstance(architectures, list)

# def test_get_ml_models(client):
#     models = client.get_ml_models(workspaceId=client.workspace)
#     assert isinstance(models, list)

# def test_create_ml_model(client, create_ml_model_fixture):
#     modelId = create_ml_model_fixture
#     assert isinstance(modelId, str)
#     models = client.get_ml_models(modelId=modelId, workspaceId=client.workspace)
#     assert len(models) == 1
#     assert models[0]['status'] == 'success'

# @pytest.mark.depends(['test_create_ml_model', 'test_edit_ml_model', 'test_download_ml_model', 'test_delete_ml_inference'])
# def test_delete_ml_model(client, create_ml_model_fixture):
#     modelId = create_ml_model_fixture
#     status = client.delete_ml_model(modelId=modelId, workspaceId=client.workspace)
#     assert status == True

# def test_edit_ml_model(client, create_ml_model_fixture):
#     modelId = create_ml_model_fixture
#     status = client.edit_ml_model(modelId=modelId, name='anatools-ml-model-edit', workspaceId=client.workspace)
#     assert status == True

# def test_download_ml_model(download_ml_model_fixture):
#     modelFile = download_ml_model_fixture
#     assert os.path.isfile(modelFile)

# def test_upload_ml_model(client, download_ml_model_fixture):
#     modelFile = download_ml_model_fixture
#     modelId = client.upload_ml_model(name=f"anatools-ml-model-upload", description="anatools test upload ml model", modelFile=modelFile, workspaceId=client.workspace)
#     assert isinstance(modelId, str)
#     while client.get_ml_models(modelId=modelId, workspaceId=client.workspace)[0]['status'] not in ['success', 'failed', 'cancelled', 'timeout']: time.sleep(10)
#     models = client.get_ml_models(modelId=modelId, workspaceId=client.workspace)
#     assert len(models) == 1
#     assert models[0]['status'] == 'success'

# def test_get_ml_inferences(client):
#     inferences = client.get_ml_inferences(workspaceId=client.workspace)
#     assert isinstance(inferences, list)

# def test_get_ml_inference_metrics(client, test_create_ml_inference):
#     inferenceId = test_create_ml_inference
#     metrics = client.get_ml_inference_metrics(inferenceId=inferenceId, workspaceId=client.workspace)
#     assert isinstance(metrics, list)

# def test_create_ml_inference(client, create_ml_inference_fixture):
#     inferenceId = create_ml_inference_fixture
#     assert isinstance(inferenceId, str)
#     inferences = client.get_ml_inferences(inferenceId=inferenceId, workspaceId=client.workspace)
#     assert len(inferences) == 1
#     assert inferences[0]['status'] == 'success'    

# @pytest.mark.dependency(depends=["test_create_ml_inference", "test_download_ml_inference"])
# def test_delete_ml_inference(client, create_ml_inference_fixture):
#     inferenceId = create_ml_inference_fixture
#     status = client.delete_ml_inference(inferenceId=inferenceId, workspaceId=client.workspace)
#     assert status == True

# def test_download_ml_inference(client, create_ml_inference_fixture):
#     inferenceId = create_ml_inference_fixture
#     inferenceFile = client.download_ml_inference(inferenceId=inferenceId, workspaceId=client.workspace)
#     assert os.path.isfile(inferenceFile)
#     os.remove(inferenceFile)
    


"""
Annotations Functions
"""


def get_annotation_formats(self):
    """Retrieves the annotation formats supported by the Rendered.ai Platform.
    
    Returns
    -------
    list[str]
        The annotation formats supported.
    """
    if self.check_logout(): return
    return self.ana_api.getAnnotationFormats()


def get_annotation_maps(self, organizationId=None, workspaceId=None, mapId=None, cursor=None, limit=None, filters=None, fields=None):
    """Retrieves annotation map information. If neither organizationId or workspaceId are specified, it will use the current workspace.
    
    Parameters
    ----------
    organizationId : str
        Organization ID to retrieve maps for.
    workspaceId: str
        Workspace ID to retrieve maps for.
    mapId: str
        Annotation map ID to retrieve.
    cursor: str
        Cursor for pagination.
    limit: int
        Maximum number of maps to return.
    filters: dict
        Filters that limit output to entries that match the filter
    fields: list[str]
        List of fields to return, leave empty to get all fields.
    
    Returns
    -------
    list[dict]
        The requested annotation maps.
    """
    self.check_logout()
    if organizationId is None and workspaceId is None: workspaceId = self.workspace
    if limit is not None and limit <= 100: items = limit
    else: items = 100
    maps = []
    while True:
        if limit and len(maps) + items > limit: items = limit - len(maps)
        ret = self.ana_api.getAnnotationMaps(organizationId=organizationId, workspaceId=workspaceId, mapId=mapId, limit=items, cursor=cursor, filters=filters, fields=fields)
        maps.extend(ret)
        if len(ret) < items or len(maps) == limit: break
        cursor = ret[-1]["mapId"]
    return maps


def get_annotations(self, datasetId=None, annotationId=None, workspaceId=None, cursor=None, limit=None, filters=None, fields=None):
    """Retrieve information about existing annotations generated for a dataset.
    
    Parameters
    ----------
    datasetId : str
        Dataset ID to generate annotations for.
    annotationId : str
        Annotation ID for a specific annotations job.
    workspaceId: str
        Workspace ID where the annotations exist. If none is provided, the current workspace will get used.
    cursor: str
        Cursor for pagination.
    limit: int
        Maximum number of annotations to return.
    filters: dict
        Filters that limit output to entries that match the filter.
    fields: list[str]
        List of fields to return, leave empty to get all fields.
    
    Returns
    -------
    list[dict]
        Annotation information.
    """
    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    if limit is not None and limit <= 100: items = limit
    else: items = 100
    annotations = []
    while True:
        if limit and len(annotations) + items > limit: items = limit - len(annotations)
        ret = self.ana_api.getAnnotations(workspaceId=workspaceId, datasetId=datasetId, annotationId=annotationId, limit=items, cursor=cursor, filters=filters, fields=fields)
        annotations.extend(ret)
        if len(ret) < items or len(annotations) == limit: break
        cursor = ret[-1]["annotationId"]
    return annotations
    

def create_annotation(self, datasetId, format, mapId=None, tags=None, workspaceId=None):
    """Generates annotations for an existing dataset. 
    
    Parameters
    ----------
    datasetId : str
        Dataset ID to generate annotation for.
    format : str
        Annotation format to use.
    mapId: str
        The ID of the map file used for annotations.
    tags: list[str]
        Tags to apply to the annotation.
    workspaceId: str
        Workspace ID of the dataset to generate annotation for. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    str
        The annotationsId for the annotation job.
    """
    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    if datasetId is None: raise ValueError("The datasetId parameter is required!")
    if format is None: raise ValueError("The format parameter is required!")
    if tags is None: tags = []
    return self.ana_api.createAnnotation(workspaceId=workspaceId, datasetId=datasetId, format=format, mapId=mapId, tags=tags)
    

def download_annotation(self, annotationId, workspaceId=None):
    """Downloads annotations archive.
    
    Parameters
    ----------
    datasetId : str
        Dataset ID to download image annotation for.
    annotationId : str
        Id of previously generated image annotation. 
    workspaceId: str
        Workspace ID of the dataset to generate annotation for. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    str
        The name of the archive file that got downloaded.
    """
    import requests

    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    if annotationId is None: raise ValueError("The annotationId parameter is required!")
    url = self.ana_api.downloadAnnotation(workspaceId=workspaceId, annotationId=annotationId)
    fname = url.split('?')[0].split('/')[-1]
    with requests.get(url, stream=True) as downloadresponse:
        with open(fname, 'wb') as outfile:
            downloadresponse.raise_for_status()
            outfile.write(downloadresponse.content)
            with open(fname, 'wb') as f:
                for chunk in downloadresponse.iter_content(chunk_size=8192):
                    f.write(chunk)
    return fname


def delete_annotation(self, annotationId, workspaceId=None):
    """Delete a dataset annotation.
    
    Parameters
    ----------
    annotationId : str
        AnnoationId of the annotation job.
    workspaceId: str
        Workspace ID of the dataset to generate annotation for. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    bool
        If true, successfully deleted the annotation.
    """
    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    if annotationId is None: raise ValueError("The annotationId parameter is required!")
    return self.ana_api.deleteAnnotation(workspaceId=workspaceId, annotationId=annotationId)


def edit_annotation(self, annotationId, tags=None, workspaceId=None):
    """Edits annotations for a dataset.
    
    Parameters
    ----------
    annotationId : str
        Annotation ID for the annotation to edit. 
    tags : list[str]
        Tags for the annotation job.
    workspaceId: str
        Workspace ID where the annotation exist. If none is provided, the current workspace will get used. 
    
    Returns
    -------
    bool
        If true, successfully edited the annotation.
    """
    self.check_logout()
    if annotationId is None: raise ValueError("The annotationId parameter is required!")
    if not isinstance(tags, list): raise ValueError("The tags parameter must be a list!")
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.editAnnotation(workspaceId=workspaceId, annotationId=annotationId, tags=tags)


def upload_annotation_map(self, mapfile, name, description=None, tags=None, organizationId=None):
    """Uploades an annotation map to the microservice. The map will be owned by the specified organization.
    
    Parameters
    ----------
    mapfile : str
        The file of the annotation map - relative to the local directry.
    name : str
        A name for model.
    description : str
        Details about the model.
    tags : list[str]
        Tags for the model.
    organizationId : str
        Id of organization that owns the model, that of the anaclient if not given.
    
    Returns
    -------
    str
        The mapId for the uploaded annotation map.
    """
    import os
    from anatools.anaclient.helpers import multipart_upload_file

    self.check_logout()
    if mapfile is None: raise ValueError("The mapfile parameter is required!")
    if description is None: description = ''
    if organizationId is None: organizationId = self.organization
    if not os.path.exists(mapfile): raise Exception(f'Could not find file: {mapfile}')
    filesize = os.path.getsize(mapfile)
    fileinfo = self.ana_api.uploadAnnotationMap(organizationId=organizationId, name=name, size=filesize, description=description, tags=tags)
    mapId = fileinfo['mapId']
    parts = multipart_upload_file(mapfile, fileinfo["partSize"], fileinfo["urls"], f"Uploading annotation map {mapfile}")
    finalize_success = self.ana_api.uploadAnnotationMapFinalizer(uploadId=fileinfo['uploadId'], key=fileinfo['key'], parts=parts)
    if not finalize_success: raise Exception(f"Failed to upload mapfile {mapfile}.")
    if self.interactive: print(f"\x1b[1K\rUpload completed successfully!", flush=True)
    return mapId


def edit_annotation_map(self, mapId, name=None, description=None, tags=None):
    """Edits the name, description or tags of a map file.
    
    Parameters
    ----------
    mapId: str
        The mapId that will be updated.
    name : str
        The new name of the annotation map. Note: this name needs to be unique per organization.
    description : str
        Description of the annotation map.
    tags : list[str]
        Tags to apply to the map.
    
    Returns
    -------
    bool
        Returns True if the map was edited.
    """
    self.check_logout()
    if mapId is None: raise Exception('The mapId parameter is required!')
    if name is None and description is None and tags is None: return True
    return self.ana_api.editAnnotationMap(mapId=mapId, name=name, description=description, tags=tags)


def delete_annotation_map(self, mapId):
    """Deletes the annotation map.
    
    Parameters
    ----------
    mapId : str
        The ID of a specific Map to delete.
    
    Returns
    -------
    bool
        Returns True if the map was deleted.
    """
    self.check_logout()
    if mapId is None: raise Exception('The mapId parameter is required!')
    return self.ana_api.deleteAnnotationMap(mapId=mapId)


def download_annotation_map(self, mapId, localDir=None):
    """Download the annotation map file from your organization.
    
    Parameters
    ----------
    mapId : str
       MapId to download.
    localDir : str
        Path for where to download the annotation map. If none is provided, current working directory will be used.
    
    Returns
    -------
    str
        The name of the map file that got downloaded.
    """
    from anatools.lib.download import download_file
    import os

    self.check_logout()
    if mapId is None: raise Exception('The mapId parameter is required!')
    if localDir is None: localDir = os.getcwd()
    url = self.ana_api.downloadAnnotationMap(mapId=mapId)
    fname = url.split('?')[0].split('/')[-1]
    return download_file(url=url, fname=fname, localDir=localDir) 
    


"""
Rules Functions
"""

def get_platform_rules(self):
    """Get the rules for an organization.
    Returns
    -------
    str
        String of rules for the platform."""
    self.check_logout()
    return self.ana_api.getPlatformRules()


def get_organization_rules(self, organizationId=None):
    """Get the rules for an organization.
    
    Parameters
    ----------
    organizationId : str
        Organization ID of the organization to get the rules for. If not specified, the current organization is used.
    
    Returns
    -------
    str
        String of rules for the organization."""
    self.check_logout()
    if organizationId is None: organizationId = self.organization
    return self.ana_api.getOrganizationRules(organizationId=organizationId)


def get_workspace_rules(self, workspaceId=None):
    """Get the rules for a workspace.
    
    Parameters
    ----------
    workspaceId : str
        Workspace ID of the workspace to get the rules for. If not specified, the current workspace is used.
    
    Returns
    -------
    str
        String of rules for the workspace."""
    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    return self.ana_api.getWorkspaceRules(workspaceId=workspaceId)


def get_service_rules(self, serviceId=None):
    """Get the rules for a service.
    
    Parameters
    ----------
    serviceId : str
        Service ID of the service to get the rules for. If not specified, the current service is used.
    
    Returns
    -------
    str
        String of rules for the service."""
    self.check_logout()
    if serviceId is None: serviceId = self.service
    return self.ana_api.getServiceRules(serviceId=serviceId)


def get_user_rules(self):
    """Get the rules for the current user."""
    self.check_logout()
    return self.ana_api.getUserRules()
    

def edit_organization_rules(self, rules, organizationId=None):
    """Edit the rules for an organization.
    
    Parameters
    ----------
    organizationId : str
        Organization ID of the organization to edit the rules for. If not specified, the current organization is used.
    rules : str
        String of rules to edit for the organization."""
    self.check_logout()
    if organizationId is None: organizationId = self.organization
    if type(rules) is not str: raise Exception('Rules parameter must be a string.')
    return self.ana_api.editOrganizationRules(organizationId=organizationId, rules=rules)
    

def edit_workspace_rules(self, rules, workspaceId=None):
    """Edit the rules for a workspace.
    
    Parameters
    ----------
    workspaceId : str
        Workspace ID of the workspace to edit the rules for. If not specified, the current workspace is used.
    rules : str
        String of rules to edit for the workspace."""
    self.check_logout()
    if workspaceId is None: workspaceId = self.workspace
    if type(rules) is not str: raise Exception('Rules parameter must be a string.')
    return self.ana_api.editWorkspaceRules(workspaceId=workspaceId, rules=rules)


def edit_service_rules(self, rules, serviceId):
    """Edit the rules for a service.
    
    Parameters
    ----------
    serviceId : str
        Service ID of the service to edit the rules for. If not specified, the current service is used.
    rules : str
        String of rules to edit for the service."""
    self.check_logout()
    if type(rules) is not str: raise Exception('Rules parameter must be a string.')
    return self.ana_api.editServiceRules(serviceId=serviceId, rules=rules)
    

def edit_user_rules(self, rules):
    """Edit the rules for the current user.
    
    Parameters
    ----------
    rules : str
        String of rules to edit for the user."""
    self.check_logout()
    if type(rules) is not str: raise Exception('Rules parameter must be a string.')
    return self.ana_api.editUserRules(rules=rules)


# Copyright 2019-2022 DADoES, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License in the root directory in the "LICENSE" file or at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import glob

logger = logging.getLogger(__name__)

class DirectoryObject:
    """ A directory stored on a volume """
    def __init__(self, directory):
        self.directory = directory

    def get_files(self, exclude_anameta = True, exclude_subdirs = True):
        """
        Get a list of the files in the directory. By default
        subdirectories and files ending in .anameta are excluded
        """
        path = os.path.join(self.directory, "*")
        files = glob.glob(path)
        path = os.path.join(self.directory, "*.anameta")
        anameta_files = glob.glob(path)
        if exclude_anameta:
            files = list(set(files) - set(anameta_files))
        if exclude_subdirs:
            for file in files[:]:
                if os.path.isdir(file):
                    files.remove(file)
        return files

    def toJSON(self):
        return {
            "class": self.__class__.__name__,
            "directory": self.directory
        }



def test_get_data_types(client):
    datasets = client.get_data_types()
    print(datasets)
    assert isinstance(datasets, list)
    assert len(datasets) > 0

def test_get_data_fields(client):
    fields = client.get_data_fields(type='Dataset')
    print(fields)
    assert isinstance(fields, list)
    assert len(fields) > 0
    assert 'datasetId' in fields



import pytest


### Fixtures ###
@pytest.fixture(scope="session")
def create_workspace_fixture(client):
    workspaceId = client.create_workspace(name=f"anatools-workspace-{client.timestamp}", organizationId=client.organization)
    return workspaceId


### Tests ###
def test_set_workspace(client):
    workspaceId = client.workspace
    status = client.set_workspace(workspaceId=workspaceId)
    assert status == True
    assert client.workspace == workspaceId

def test_get_workspaces(client, create_workspace_fixture):
    workspaces = client.get_workspaces()
    assert isinstance(workspaces, list)
    assert len(workspaces) > 0
    assert create_workspace_fixture in [w['workspaceId'] for w in workspaces]

@pytest.mark.dependency(depends=["test_edit_workspace"])
def test_delete_workspace(client, create_workspace_fixture):
    workspaceId = create_workspace_fixture
    status = client.delete_workspace(workspaceId=workspaceId)
    assert status == True

def test_edit_workspace(client, create_workspace_fixture):
    workspaceId = create_workspace_fixture
    status = client.edit_workspace(workspaceId=workspaceId, name=f"anatools-workspace-edit-{client.timestamp}", channelIds=[client.channel])
    assert status == True